"""
Use the generic "signature" tree in the workspace to brute
force attempt to find function entry points.  This is
slightly desperate, so do it late when most locations are
already defined...  Additionally, if the "pointers" generic
module is run first, there is a reasonabily high likelyhood
that the code this finds is dead...
"""
import logging
import traceback

import envi
import envi.memory as e_mem
import envi.const as e_const
import vivisect

logger = logging.getLogger(__name__)


def analyze(vw):
    """
    Assuming that a bunch of functions have already been defined and
    fully analyzed, use the compiler against itself and attempt to
    brute force find other function entry points based on the
    entry signatures db.
    """
    # FIXME make the below code an undefined space iterator
    # and use in findPointers too
    for mapva, mapsize, mapflags, fname in vw.getMemoryMaps():

        # Segment permissions check for likely code stuff at all
        if not mapflags & e_const.MM_EXEC:
            continue
        if mapflags & e_const.MM_UNINIT:
            continue

        i = 0
        maxsize = mapsize - 4
        while i < maxsize:
            va = mapva + i
            loctup = vw.getLocation(va)
            if loctup is not None:
                i += loctup[vivisect.L_SIZE]
                continue

            i += 1

            try:

                if vw.isFunctionSignature(va):
                    logger.debug('discovered new function (by signature): 0x%x', va)
                    vw.makeFunction(va)

            except vivisect.InvalidLocation as msg:
                logger.error("InvalidLocation: %s", msg)
            except envi.InvalidInstruction:
                continue
            except envi.EnviException as msg:
                logger.error("%s: %s" % (msg.__class__.__name__, msg))
            except Exception:
                logger.error(traceback.format_exc())
                continue



"""
The H8 Emulator module.
"""

import struct
import logging

import envi
import envi.bits as e_bits
import envi.const as e_const
import envi.archs.h8.regs as h8_regs
import envi.archs.h8.const as h8_const

from envi.archs.h8 import H8Module
from envi.archs.h8.operands import H8RegDirOper


logger = logging.getLogger(__name__)

# bcd lookup table generation
bcd_add = {}
for x in range(0xa):
    for y in range(0xa):
        bcd_add[(0, 0, (x, y))] = 0, 0

    for y in range(4):
        bcd_add[(0, 1, (x, y))] = 0x6, 0

for x in range(9):
    for y in range(0xa, 0x10):
        bcd_add[(0, 0, (x, y))] = 0x6, 0

for x in range(0xa, 0x10):
    for y in range(0xa):
        bcd_add[(0, 0, (x, y))] = 0x60, 1

    for y in range(4):
        bcd_add[(0, 1, (x, y))] = 0x66, 1

for x in range(9, 0x10):
    for y in range(0xa, 0x10):
        bcd_add[(0, 0, (x, y))] = 0x66, 1

for x in range(3):
    for y in range(0xa):
        bcd_add[(1, 0, (x, y))] = 0x60, 1

    for y in range(0xa, 0x10):
        bcd_add[(1, 0, (x, y))] = 0x66, 1

for x in range(4):
    for y in range(4):
        bcd_add[(1, 1, (x, y))] = 0x66, 1

bcd_sub = {}
for x in range(0xa):
    for y in range(0xa):
        bcd_sub[(0, 0, (x, y))] = 0, 0

for x in range(0x9):
    for y in range(6, 0x10):
        bcd_sub[(0, 1, (x, y))] = 0xfa, 0

for x in range(7, 0x10):
    for y in range(0xa):
        bcd_sub[(1, 0, (x, y))] = 0xa0, 1

for x in range(6, 0x10):
    for y in range(6, 0x10):
        bcd_sub[(1, 1, (x, y))] = 0x9a, 1

sign_ext_bits = {
        1: [((-1 & 0xff) << (8 - x) & 0xff) for x in range(8)],
        2: [],
        4: [],
        }


# calling conventions
class H8CallAdv(envi.CallingConvention):
    """
    H8 architectures hand arguments between functions using

    reference: C_H8_User_Manual.pdf
    """
    arg_def = [(e_const.CC_REG, h8_regs.REG_ER0),
               (e_const.CC_REG, h8_regs.REG_ER1),
               (e_const.CC_REG, h8_regs.REG_ER2),
               (e_const.CC_STACK_INF, 4)]
    retaddr_def = (e_const.CC_STACK_INF, 0)
    retval_def = (e_const.CC_REG, h8_regs.REG_ER0), (e_const.CC_STACK_INF, 0)  # FIXME
    flags = e_const.CC_CALLEE_CLEANUP
    align = 2
    pad = 0

    def execCallReturn(self, emu, value, argc=None):
        sp = emu.getRegister(h8_regs.REG_SP)
        pc = struct.unpack('>H', emu.readMemory(sp, 2))[0]
        sp += 2  # For the saved pc
        sp += (2 * argc)  # Cleanup saved args

        emu.setRegister(h8_regs.REG_SP, sp)
        emu.setRegister(h8_regs.REG_R0, value)
        emu.setProgramCounter(pc)

    def getCallArgs(self, emu, count):
        # r0-r3 are used to hand in parameters.  additional ph8s are stored and pointed to by r0
        return emu.getRegister(0xf)


h8call = H8CallAdv()

CPUSTATE_RESET = 0
CPUSTATE_EXC = 1
CPUSTATE_EXEC = 2
CPUSTATE_BUS = 3
CPUSTATE_SLEEP = 4
CPUSTATE_SWSTDBY = 5
CPUSTATE_HWSTDBY = 6


class H8Emulator(H8Module, h8_regs.H8RegisterContext, envi.Emulator):
    IVT_RESET = 0

    def __init__(self, advanced=True):
        H8Module.__init__(self)
        envi.Emulator.__init__(self, self)
        h8_regs.H8RegisterContext.__init__(self)

        self.state = CPUSTATE_RESET
        self.ptrsz = 0

        # seglist = [(0,0xffffffff) for x in range(6)]

        self.setAdvanced(advanced)
        self.addCallingConvention("H8 Arch Procedure Call", h8call)

    def setAdvanced(self, advanced=True):
        self.advanced = advanced
        if advanced:
            self.ptrsz = 4
        else:
            self.ptrsz = 2

    def reset(self):
        '''
        triggers the reset interrupt
        '''
        if self.getProgramCounter() == 0:
            va = self.readMemoryPtr(self.IVT_RESET)
            self.setProgramCounter(va)

    def getPointerSize(self):
        return self.ptrsz

    def processInterrupt(self, intval=0):
        '''
        From outside the emulator, cause an interrupt to be handled.
        FIXME: wants to pause the run()/runFunction() thread, if any active
        '''
        va = self.emuProcessInterrupt_2140(intval)
        self.setProgramCounter(va)

    def emuProcessInterrupt_2140(self, intval=0):
        '''
        interrupt processing done according to H8/300 standards.  subclass and modify this function
        if the desired behavior differs.
        supports: advanced and normal modes
        '''
        if self.getFlag(CCR_I):
            # FIXME: check for NMI or RESET
            if intval != 0 and intval not in NMI_INTS:
                logger.info("Ignoring maskable interrupt (CCR_I flag is set)")
                return

        logger.info("Interrupt Handler: 0x%x", intval)

        addr = 0

        if self.advanced:
            if intval != 0:
                val = (self.getStatusRegister() << 24) | self.getStatusRegister()
                self.doPush(val, 4)

                addr = (intval * 4) + 0xc

            isrAddr = self.readMemortPtr(addr)
        else:
            if intval != 0:
                val = self.getStatusRegister()
                val |= (val << 8)
                self.doPush(val, 2)
                self.doPush(self.getProgramCounter())

                addr = (intval * 2) + 6

            isrAddr = self.readMemoryPtr(addr)

        return isrAddr

    def emuProcessInterrupt(self, intval=0):
        '''
        interrupt processing done according to H8/300 standards.  subclass and modify this function
        if the desired behavior differs.
        supports: advanced and normal modes
        '''
        logger.info("Interrupt Handler: 0x%x", intval)

        addr = 0

        if self.advanced:
            if intval != 0:
                self.doPush(self.getStatusRegister())
                self.doPush(self.getStatusRegister())
                self.doPush(self.getProgramCounter())

                addr = (intval * 4) + 0xa

            isrAddr = self.readMemortPtr(addr)
        else:
            if intval != 0:
                self.doPush(self.getStatusRegister())
                self.doPush(self.getProgramCounter())

                addr = (intval * 2) + 4

            isrAddr = self.readMemoryPtr(addr)

        return isrAddr

    def undefFlags(self):
        """
        Used in PDE.
        A flag setting operation has resulted in un-defined value.  Set
        the flags to un-defined as well.
        """
        self.setRegister(h8_const.REG_FLAGS, None)

    def setFlag(self, which, state):
        flags = self.getRegister(h8_const.REG_FLAGS)
        if state:
            flags |= which
        else:
            flags &= ~which
        self.setRegister(h8_const.REG_FLAGS, flags)

    def getFlag(self, which):
        flags = self.getRegister(h8_const.REG_FLAGS)
        if flags is None:
            raise envi.PDEUndefinedFlag(self)
        return bool(flags & which)

    def executeOpcode(self, op):
        # NOTE: If an opcode method returns
        #       other than None, that is the new pc
        x = None
        meth = self.op_methods.get(op.mnem, None)
        if meth is None:
            raise envi.UnsupportedInstruction(self, op)
        x = meth(op)

        if x is None:
            pc = self.getProgramCounter()
            x = pc+op.size

        self.setProgramCounter(x)

    def doPush(self, val, inc=2, reg=h8_const.REG_SP):
        sp = self.getRegister(reg)
        sp -= inc
        self.writeMemValue(sp, val, inc)
        self.setRegister(reg, sp)

    def doPop(self, inc=2, reg=h8_const.REG_SP):
        sp = self.getRegister(reg)
        val = self.readMemValue(sp, inc)
        self.setRegister(reg, sp+inc)
        return val

    def i_and(self, op):
        dsize = op.opers[1].tsize
        res = self.logicalAnd(op)
        self.setOperValue(op, 1, res)

        self.setFlag(h8_regs.CCR_Z, not res)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(res, dsize))
        self.setFlag(h8_regs.CCR_V, 0)

    def i_andc(self, op):
        res = self.logicalAnd(op)
        self.setOperValue(op, 1, res)

    def i_band(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        val = self.getOperValue(op, 1)

        val >>= bit
        val &= C

        self.setFlag(h8_regs.CCR_C, val)

    def i_bra(self, op):
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_brn(self, op):
        pass

    def i_bhi(self, op):
        if not (self.getFlag(h8_regs.CCR_C) == 0 or self.getFlag(h8_regs.CCR_Z) == 0):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bls(self, op):
        if not (self.getFlag(h8_regs.CCR_C) or self.getFlag(h8_regs.CCR_Z)):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bhs(self, op):
        if self.getFlag(h8_regs.CCR_C):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_blo(self, op):
        if not self.getFlag(h8_regs.CCR_C):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bne(self, op):
        if self.getFlag(h8_regs.CCR_Z):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_beq(self, op):
        if not self.getFlag(h8_regs.CCR_Z):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bvc(self, op):
        if self.getFlag(h8_regs.CCR_V):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bvs(self, op):
        if not self.getFlag(h8_regs.CCR_V):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bpl(self, op):
        if self.getFlag(h8_regs.CCR_N):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bmi(self, op):
        if not self.getFlag(h8_regs.CCR_N):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bge(self, op):    # FIXME: TEST.  these last 4 seem mixed up.
        if self.getFlag(h8_regs.CCR_V) != self.getFlag(h8_regs.CCR_N):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_blt(self, op):    # FIXME: TEST.  these last 4 seem mixed up.
        if self.getFlag(h8_regs.CCR_V) == self.getFlag(h8_regs.CCR_N):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_bgt(self, op):    # FIXME: TEST.  these last 4 seem mixed up.
        if (self.getFlag(h8_regs.CCR_V) != self.getFlag(h8_regs.CCR_N)) or self.getFlag(h8_regs.CCR_Z):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    def i_ble(self, op):    # FIXME: TEST.  these last 4 seem mixed up.
        if not ((self.getFlag(h8_regs.CCR_V) != self.getFlag(h8_regs.CCR_N)) and self.getFlag(h8_regs.CCR_Z)):
            return
        nextva = self.getOperValue(op, 0)
        return nextva

    i_bt = i_bra
    i_bf = i_brn
    i_bcc = i_bhs
    i_bcs = i_blo

    def i_bclr(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        tgt &= ~(1 << bit)
        self.setOperValue(op, 1, tgt)

    def i_biand(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C &= ~(tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bild(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C = ~(tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bior(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C |= ~(tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bist(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        tgt &= ~(1 << bit)
        tgt |= (C ^ 1) << bit
        self.setOperValue(op, 1, tgt)

    def i_bixor(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C ^= ~(tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bld(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C = (tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bnot(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        tgt ^= (1 << bit)
        self.setOperValue(op, 1, tgt)

    def i_bor(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C |= (tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_bset(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        tgt |= (1 << bit)
        self.setOperValue(op, 1, tgt)

    def i_bst(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        tgt &= ~(1 << bit)
        tgt |= C << bit
        self.setOperValue(op, 1, tgt)

    def i_btst(self, op):
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        Z = ((tgt >> bit) & 1) ^ 1
        self.setFlag(h8_regs.CCR_Z, Z)

    def i_bxor(self, op):
        C = self.getFlag(h8_regs.CCR_C)
        bit = self.getOperValue(op, 0)
        tgt = self.getOperValue(op, 1)

        C ^= (tgt >> bit) & 1
        self.setFlag(h8_regs.CCR_C, C)

    def i_cmp(self, op):
        flagtup = self.integerSubtraction(op)
        self.doFlags(flagtup)

    def i_bsr(self, op):
        nextva = self.getProgramCounter()

        if self.advanced:
            self.doPush(nextva, 4)
        else:
            self.doPush(nextva)

        disp = self.getOperValue(op, 0)
        pc = self.getProgramCounter() + disp
        return pc

    def i_jsr(self, op):
        nextva = self.getProgramCounter()

        if self.advanced:
            self.doPush(nextva, 4)
        else:
            self.doPush(nextva)

        ea = self.getOperValue(op, 0)
        return ea

    def i_stm(self, op):
        reglist = self.getOperValue(op, 0)

        for regval in reglist:
            op.opers[1].setOperValue(op, self, regval)

    def i_ldm(self, op):
        reglist = op.opers[1].getOperRegs(op)

        for reg in reglist:
            regval = op.opers[0].getOperValue(op, self, mod=True)
            self.setRegister(reg, regval)

    def i_mov(self, op):
        ssize = op.opers[0].tsize
        val = self.getOperValue(op, 0)
        self.setOperValue(op, 1, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, ssize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)

        if isinstance(op.opers[1], H8RegDirOper) and op.opers[1].reg == h8_const.REG_PC:
            return val

    def i_add(self, op):
        (ssize, dsize, sres, ures, sdst, udst) = self.integerAddition(op)

        self.setOperValue(op, 1, ures)

        # FIXME: test and validate
        self.setFlag(h8_regs.CCR_H, e_bits.is_signed_half_carry(sres, dsize, sdst))
        self.setFlag(h8_regs.CCR_C, e_bits.is_unsigned_carry(ures, dsize))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(sres, dsize))

    def i_adds(self, op):
        (ssize, dsize, sres, ures, sdst, udst) = self.integerAddition(op)

        self.setOperValue(op, 1, ures)

    def i_addx(self, op):
        (ssize, dsize, sres, ures, sdst, udst) = self.integerAddition(op)

        C = self.getFlag(h8_regs.CCR_C)
        sres += C
        ures += C

        self.setOperValue(op, 1, ures)

        # FIXME: test and validate  (same as i_add)
        self.setFlag(h8_regs.CCR_H, e_bits.is_signed_half_carry(sres, dsize, sdst))
        self.setFlag(h8_regs.CCR_C, e_bits.is_unsigned_carry(ures, dsize))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(sres, dsize))

    def i_daa(self, op):
        oper = self.getOperValue(op, 0)
        upop = oper >> 4 & 0xf
        loop = oper & 0xf
        C = self.getFlag(h8_regs.CCR_C)
        H = self.getFlag(h8_regs.CCR_H)

        addtup = bcd_add.get((C, H, upop, loop))
        if addtup is None:
            logger.debug("DAA:  %x %x %x %x - addtup is None", C, H, upop, loop)
            return  # FIXME: raise exception once figured out
        addval, resC = addtup
        ures = addval + oper

        self.setOperValue(op, 0, ures)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, 1))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_C, resC)

    def i_das(self, op):
        oper = self.getOperValue(op, 0)
        upop = oper >> 4 & 0xf
        loop = oper & 0xf
        C = self.getFlag(h8_regs.CCR_C)
        H = self.getFlag(h8_regs.CCR_H)

        addtup = bcd_sub.get((C, H, upop, loop))
        if addtup is None:
            logger.debug("DAS:  %x %x %x %x - addtup is None", C, H, upop, loop)
            # FIXME: raise exception once figured out
            return
        addval, resC = addtup
        ures = addval + oper

        self.setOperValue(op, 0, ures)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, 1))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_C, resC)       # this should always be what it was coming in

    def i_inc(self, op):
        dstidx = len(op.opers) - 1
        if dstidx == 1:
            ssize = op.opers[0].tsize
            dsize = op.opers[1].tsize
            src = self.getOperValue(op, 0)
            dst = self.getOperValue(op, 1)

            udst = e_bits.unsigned(dst, dsize)
            # TODO: What is sdst and why does it exist?
            # sdst = e_bits.signed(dst, dsize)
            usrc = e_bits.unsigned(src, ssize)
            ssrc = e_bits.signed(src, ssize)

        else:
            dsize = op.opers[0].tsize
            dst = self.getOperValue(op, 0)

            udst = e_bits.unsigned(dst, dsize)
            # sdst = e_bits.signed(dst, dsize)
            ssrc = usrc = 1

        ures = usrc + udst
        sres = ssrc + udst

        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(sres, dsize))
        # V must be set if previous value was 0x7f (per docs, page 78 of H8/300)

        self.setOperValue(op, dstidx, ures)

    def i_dec(self, op):
        dstidx = len(op.opers) - 1
        if dstidx == 1:
            ssize = op.opers[0].tsize
            dsize = op.opers[1].tsize
            src = self.getOperValue(op, 0)
            dst = self.getOperValue(op, 1)

            udst = e_bits.unsigned(dst, dsize)
            sdst = e_bits.signed(dst, dsize)
            usrc = e_bits.unsigned(src, ssize)
            ssrc = e_bits.signed(src, ssize)

        else:
            dsize = op.opers[0].tsize
            dst = self.getOperValue(op, 0)

            udst = e_bits.unsigned(dst, dsize)
            sdst = e_bits.signed(dst, dsize)
            ssrc = usrc = 1

        ures = udst - usrc
        sres = sdst - ssrc

        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(sres, dsize))
        # V must be set if previous value was 0x80 (per docs, page 73 of H8/300)

        self.setOperValue(op, dstidx, ures)

    def i_jmp(self, op):
        return self.getOperValue(op, 0)

    def i_sub(self, op):
        # Src op gets sign extended to dst
        flagtup = self.integerSubtraction(op)
        self.doFlags(flagtup)
        (ssize, dsize, sres, ures, sdst, udst) = flagtup
        self.setOperValue(op, 1, ures)

    def doFlags(self, flagtup):
        (ssize, dsize, sres, ures, sdst, udst) = flagtup

        self.setFlag(h8_regs.CCR_H, e_bits.is_signed_half_carry(ures, dsize, udst))
        self.setFlag(h8_regs.CCR_C, e_bits.is_unsigned_carry(ures, dsize))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(sres, dsize))

    def i_subs(self, op):
        # Src op gets sign extended to dst
        ssize = op.opers[0].tsize
        # TODO: Why?
        # dsize = op.opers[1].tsize
        src = e_bits.sign_extend(self.getOperValue(op, 0), ssize, self.ptrsz)
        dst = e_bits.sign_extend(self.getOperValue(op, 1), ssize, self.ptrsz)

        if src is None or dst is None:
            self.undefFlags()
            return None

        res = src - dst
        self.setOperValue(op, 1, res)

    def i_subx(self, op):
        # Src op gets sign extended to dst
        ssize = op.opers[0].tsize
        dsize = op.opers[1].tsize
        src = e_bits.sign_extend(self.getOperValue(op, 0), ssize, self.ptrsz)
        dst = e_bits.sign_extend(self.getOperValue(op, 1), dsize, self.ptrsz)
        C = self.getFlag(h8_regs.CCR_C)

        if src is None or dst is None:
            self.undefFlags()
            return None

        (ssize, dsize, sres, ures, sdst, udst) = self.intSubBase(dst, src + C, dsize, ssize)
        self.setOperValue(op, 1, ures)

    def i_xor(self, op):
        src1 = self.getOperValue(op, 0)
        src2 = self.getOperValue(op, 1)

        # FIXME PDE and flags
        if src1 is None or src2 is None:
            self.undefFlags()
            self.setOperValue(op, 1, None)
            return

        usrc1 = e_bits.unsigned(src1, 4)
        usrc2 = e_bits.unsigned(src2, 4)

        ures = usrc1 ^ usrc2

        self.setOperValue(op, 1, ures)

        self.setFlag(h8_regs.CCR_C, e_bits.is_unsigned_carry(ures, 4))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, 4))
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(ures, 4))

    def i_xorc(self, op):
        src = self.getOperValue(op, 0)
        dst = self.getStatusRegister()

        # FIXME PDE and flags
        if src is None:
            self.undefFlags()
            self.setOperValue(op, 1, None)
            return

        # TODO: WHY?
        # ssize = op.opers[0].tsize
        # dsize = op.opers[1].tsize

        ures = src ^ dst

        self.setStatusRegister(ures)

    def i_nop(self, op):
        pass

    def i_divxu(self, op):
        divisor = self.getOperValue(op, 0)
        dividend = self.getOperValue(op, 1)

        quotient = int(dividend / divisor)
        remainder = dividend % divisor

        rdval = (remainder << 8) | quotient

        self.setOperValue(op, 1, rdval)

        self.setFlag(h8_regs.CCR_Z, not quotient)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(quotient, 4))

    def i_divxs(self, op):
        ssize = op.opers[0].tsize
        dsize = op.opers[1].tsize

        divisor = self.getOperValue(op, 0)
        dividend = self.getOperValue(op, 1)

        sdivisor = e_bits.signed(divisor, ssize)
        sdividend = e_bits.signed(dividend, dsize)

        quotient = int(sdividend / sdivisor)
        remainder = sdividend % sdivisor

        rdval = (remainder << 8) | quotient

        self.setOperValue(op, 1, rdval)

        self.setFlag(h8_regs.CCR_Z, not quotient)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(quotient, 4))

    def i_eepmov(self, op):
        if op.iflags & h8_const.IF_W:
            delta = 2
            count = self.getRegisterByName('r4')
        elif op.iflags & h8_const.IF_B:
            delta = 1
            count = self.getRegisterByName('r4l')

        src = self.getRegister(h8_regs.REG_ER5)
        dst = self.getRegister(h8_regs.REG_ER6)
        logger.info("WARNING: EEPMOV instruction executed... 0x%x -> 0x%x (count=0x%x)", src, dst, count * delta)

        for x in range(count):
            data = self.readMemValue(src, delta)
            self.writeMemValue(dst, data, delta)
            src += delta
            dst += delta
            # FIXME: NMI can interrupt, must be handled after the current read/write cycle

    def i_exts(self, op):
        pass

    def i_extu(self, op):
        pass

    def i_ldc(self, op):
        oper = self.getOperValue(op, 0)
        self.setStatusRegister(oper)

    def i_stc(self, op):
        oper = self.getStatusRegister()
        self.setOperValue(op, 1, oper)

    def i_movfpe(self, op):
        raise Exception("Implement MOVFPE emulator instruction with peripherals")

    def i_movtpe(self, op):
        raise Exception("Implement MOVTPE emulator instruction with peripherals")

    def i_mulxu(self, op):
        '''
        mul, extend as unsigned
        rs is 8 bits
        rd is 16 bits, but only uses the lower 8 bits for multiplicand
        product is then stored in 16-bit rd
        flags are not updated
        '''
        dsize = op.opers[1].tsize
        src = self.getOperValue(op, 0)
        dst = self.getOperValue(op, 1)

        ures = (dst * src)
        val = ures & e_bits.u_maxes[dsize]

        self.setOperValue(op, 1, val)

    def i_mulxs(self, op):
        '''
        mul, extend as signed
        rs is 8 bits
        rd is 16 bits, but only uses the lower 8 bits for multiplicand
        product is then stored in 16-bit rd
        flags are not updated
        '''
        ssize = op.opers[0].tsize
        dsize = op.opers[1].tsize
        src = self.getOperValue(op, 0)
        dst = self.getOperValue(op, 1)

        ssrc = e_bits.signed(src, ssize)
        sdst = e_bits.signed(dst, dsize)

        sres = (sdst * ssrc)
        val = sres & e_bits.u_maxes[dsize]

        self.setOperValue(op, 1, val)

    def i_neg(self, op):
        dsize = op.opers[0].tsize
        oper = self.getOperValue(op, 0)
        oper = e_bits.signed(oper, dsize)
        oper = -oper
        self.setOperValue(op, 0, oper)

        self.setFlag(h8_regs.CCR_H, e_bits.is_signed_half_carry(oper, dsize, oper))
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(oper, dsize))
        self.setFlag(h8_regs.CCR_Z, not oper)
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(oper, dsize))
        self.setFlag(h8_regs.CCR_C, e_bits.is_unsigned_carry(oper, dsize))

    def i_not(self, op):
        dsize = op.opers[0].tsize
        oper = self.getOperValue(op, 0)
        # oper = e_bits.signed(oper, dsize)
        oper = e_bits.u_maxes[dsize] - oper
        self.setOperValue(op, 0, oper)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(oper, dsize))
        self.setFlag(h8_regs.CCR_Z, not oper)
        self.setFlag(h8_regs.CCR_V, 0)

    def i_or(self, op):
        src = self.getOperValue(op, 0)
        dst = self.getOperValue(op, 1)

        # FIXME PDE and flags
        if src is None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        # TODO: WHY?
        # ssize = op.opers[0].tsize
        dsize = op.opers[1].tsize

        ures = src | dst

        self.setOperValue(op, 1, ures)
        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(ures, dsize))
        self.setFlag(h8_regs.CCR_Z, not ures)
        self.setFlag(h8_regs.CCR_V, 0)

    def i_orc(self, op):
        src = self.getOperValue(op, 0)
        dst = self.getStatusRegister()

        # FIXME PDE and flags
        if src is None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        # TODO: WHY?
        # ssize = op.opers[0].tsize
        # dsize = op.opers[1].tsize

        ures = src | dst

        self.setStatusRegister(ures)

    def i_pop(self, op):
        dsize = op.opers[0].tsize
        val = self.doPop()
        self.setOperValue(op, 0, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)

    def i_push(self, op):
        dsize = op.opers[0].tsize
        val = self.getOperValue(op, 0)
        self.doPush(val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)

    def i_rotl(self, op):
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8)
        val <<= shbits
        val |= (val >> bits)
        val &= e_bits.u_maxes[dsize]
        C = val & 1

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, C)

    def i_rotr(self, op):
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8)
        C = (val >> (shbits - 1)) & 1
        val |= (val << bits)
        val >>= shbits
        val &= e_bits.u_maxes[dsize]

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, C)

    def i_rotxl(self, op):
        '''
        rotate left, with extend Carry
        '''
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8) - 1

        C = self.getFlag(h8_regs.CCR_C)
        newC = (val >> (bits+1-shbits)) & 1
        val <<= shbits
        val |= (val >> (bits + 2))
        val |= (C << (shbits - 1))
        val &= e_bits.u_maxes[dsize]

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, newC)

    def i_rotxr(self, op):
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8)
        C = self.getFlag(h8_regs.CCR_C)
        newC = (val >> (shbits-1)) & 1
        val |= (val << (bits + 1))
        val |= (C << bits)
        val >>= shbits
        val &= e_bits.u_maxes[dsize]

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, newC)

    def i_rte(self, op):
        '''
        return from Interrupt Exception
        EXCLAMATION: stack layout for calls and ISRs differ between archs!
        '''
        logger.info("Return from Interrupt Handler")

        if self.advanced:
            pop = self.doPop(4)
            retAddr = pop & 0xffffff
            ccr = (pop >> 24) & 0xff
        else:
            ccr = self.doPop() & 0xff
            retAddr = self.doPop()

        self.setStatusRegister(ccr)
        return retAddr

    def i_rts(self, op):
        '''
        return from Subroutine
        EXCLAMATION: stack layout for calls and ISRs differ between archs!
        '''
        if self.advanced:
            pop = self.doPop(4)
            retAddr = pop & 0xffffff
        else:
            retAddr = self.doPop()

        return retAddr

    def i_shal(self, op):
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8)

        C = (val >> (bits - shbits)) & 1
        val <<= shbits
        rawval = val
        val &= e_bits.u_maxes[dsize]

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, e_bits.is_signed_overflow(rawval, dsize))
        self.setFlag(h8_regs.CCR_C, C)

    def i_shar(self, op):
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        logger.debug(repr((shbits, val, dsize)))
        max_mask = e_bits.u_maxes[dsize]
        signbits = ((-1 & max_mask) << ((8 * dsize) - shbits) & max_mask)
        S = val > e_bits.s_maxes[dsize]

        C = (val >> (shbits-1)) & 1
        val >>= shbits

        if S:
            val |= signbits

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, C)

    def i_shll(self, op):
        '''
        same as shal, except for handling of V
        '''
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        bits = (dsize * 8)

        C = (val >> (bits - shbits)) & 1
        val <<= shbits
        val &= e_bits.u_maxes[dsize]

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, C)

    def i_shlr(self, op):
        '''
        not the same as shar
        '''
        dstidx = len(op.opers) - 1
        if dstidx:
            shbits = self.getOperValue(op, 0)
            val = self.getOperValue(op, 1)
            dsize = op.opers[1].tsize
        else:
            shbits = 1
            val = self.getOperValue(op, 0)
            dsize = op.opers[0].tsize

        # bits = (dsize * 8) - 1

        C = (val >> (shbits-1)) & 1
        val >>= shbits

        self.setOperValue(op, dstidx, val)

        self.setFlag(h8_regs.CCR_N, e_bits.is_signed(val, dsize))
        self.setFlag(h8_regs.CCR_Z, not val)
        self.setFlag(h8_regs.CCR_V, 0)
        self.setFlag(h8_regs.CCR_C, C)

    def i_sleep(self, op):
        logger.info("Entering Sleep Mode... waiting for an External Interrupt")
        self.state = CPUSTATE_SLEEP

    def i_str(self, op):
        ccr = self.getStatusRegister()
        self.setOperValue(op, 0, ccr)

    def i_tas(self, op):
        dsize = op.opers[0].tsize
        logger.debug(repr(dsize))
        oper = self.getOperValue(op, 0)

        # do comparison and set flags
        flagtup = self.intSubBase(oper, 0, dsize, dsize)
        self.doFlags(flagtup)

        oper |= 0x80

        self.setOperValue(op, 0, oper)

    def i_trapa(self, op):
        logger.info("SW EXCEPTION:  %s", op)
        # FIXME: processInterrupt()



'''
Some utils for QT code which uses vstruct...
'''
import vqt.tree as vq_tree

from PyQt5.QtWidgets import *

from vqt.main import idlethreadsync

class VQStructNamespacesView(vq_tree.VQTreeView):

    def __init__(self, parent=None):
        vq_tree.VQTreeView.__init__(self, parent=parent)
        #model = vq_tree.VTreeView(parent=self, columns=('Namespace', 'Structure'))

class VQStructSelectView(vq_tree.VQTreeView):

    def __init__(self, vsbuilder, parent=None):
        vq_tree.VQTreeView.__init__(self, parent=parent)
        self.vsbuilder = vsbuilder

        model = vq_tree.VQTreeModel(parent=self, columns=('Namespace', 'Structure'))
        for nsname in vsbuilder.getVStructNamespaceNames():
            pnode = model.append((nsname, ''))
            pnode.structname = None
            for sname in vsbuilder.getVStructNames(namespace=nsname):
                spnode = model.append(('', sname), parent=pnode)
                spnode.structname = '%s.%s' % (nsname, sname)

        for sname in vsbuilder.getVStructNames():
            node = model.append( ('', sname ) )
            node.structname = sname

        self.setModel(model)

class VQStructSelectDialog(QDialog):

    def __init__(self, vsbuilder, parent=None):
        QDialog.__init__(self, parent=parent)
        self.structname = None

        self.setWindowTitle('Select a structure...')

        vlyt = QVBoxLayout()
        hlyt = QHBoxLayout()

        self.structtree = VQStructSelectView(vsbuilder, parent=self)

        hbox = QWidget(parent=self)

        ok = QPushButton("Ok", parent=hbox)
        cancel = QPushButton("Cancel", parent=hbox)

        self.structtree.doubleClicked.connect( self.dialog_activated )

        ok.clicked.connect(self.dialog_ok)
        cancel.clicked.connect(self.dialog_cancel)

        hlyt.addStretch(1)
        hlyt.addWidget(cancel)
        hlyt.addWidget(ok)
        hbox.setLayout(hlyt)

        vlyt.addWidget(self.structtree)
        vlyt.addWidget(hbox)
        self.setLayout(vlyt)

        self.resize(500, 500)

    def dialog_activated(self, idx):
        if idx.isValid():
            pnode = idx.internalPointer()
            self.structname = pnode.structname
        self.accept()

    def dialog_ok(self):
        for idx in self.structtree.selectedIndexes():
            pnode = idx.internalPointer()
            self.structname = pnode.structname
        self.accept()

    def dialog_cancel(self):
        self.reject()

@idlethreadsync
def selectStructure(vsbuilder, parent=None):
    d = VQStructSelectDialog(vsbuilder, parent=parent)
    r = d.exec_()
    return d.structname

class VQStructNamespacesView(vq_tree.VQTreeView):

    def __init__(self, parent=None):
        vq_tree.VQTreeView.__init__(self, parent=parent)

        model = vq_tree.VQTreeModel(parent=self, columns=('Subsystem', 'Module Name'))

        win = model.append(('windows', ''))
        xp_i386_user = model.append(('Windows XP i386 Userland', ''), parent=win)
        xp_i386_ntdll = model.append(('','ntdll'), parent=xp_i386_user)
        xp_i386_ntdll.modinfo = ('ntdll','vstruct.defs.windows.win_5_1_i386.ntdll')

        xp_i386_kern = model.append(('Windows XP i386 Kernel', ''), parent=win)
        xp_i386_nt = model.append(('','nt'), parent=xp_i386_kern)
        xp_i386_nt.modinfo = ('nt','vstruct.defs.windows.win_5_1_i386.ntoskrnl')
        xp_i386_win32k = model.append(('','win32k'), parent=xp_i386_kern)
        xp_i386_win32k.modinfo = ('win32k','vstruct.defs.windows.win_5_1_i386.win32k')

        win7_amd64_user = model.append(('Windows 7 amd64 Userland', ''), parent=win)
        win7_amd64_ntdll = model.append(('','ntdll'), parent=win7_amd64_user)
        win7_amd64_ntdll.modinfo = ('ntdll','vstruct.defs.windows.win_6_1_amd64.ntdll')

        pos = model.append(('posix',''))
        pos_elf = model.append(('', 'Elf'), parent=pos)
        pos_elf.modinfo = ('elf', 'vstruct.defs.elf')

        osx  = model.append(('osx', ''))
        osx_macho = model.append(('', 'macho'), parent=osx)

        self.setModel(model)

class VQStructNamespaceDialog(QDialog):

    def __init__(self, parent=None):
        QDialog.__init__(self, parent=parent)
        self.modinfo = None

        self.setWindowTitle('Select a module...')

        vlyt = QVBoxLayout()
        hlyt = QHBoxLayout()

        self.structtree = VQStructNamespacesView(parent=self)

        hbox = QWidget(parent=self)

        ok = QPushButton("Ok", parent=hbox)
        cancel = QPushButton("Cancel", parent=hbox)

        self.structtree.doubleClicked.connect( self.dialog_activated )

        ok.clicked.connect(self.dialog_ok)
        cancel.clicked.connect(self.dialog_cancel)

        hlyt.addStretch(1)
        hlyt.addWidget(cancel)
        hlyt.addWidget(ok)
        hbox.setLayout(hlyt)

        vlyt.addWidget(self.structtree)
        vlyt.addWidget(hbox)
        self.setLayout(vlyt)

        self.resize(500, 500)

    def dialog_activated(self, idx):
        if idx.isValid():
            pnode = idx.internalPointer()
            self.modinfo = getattr(pnode, 'modinfo', None)
        self.accept()

    def dialog_ok(self):
        for idx in self.structtree.selectedIndexes():
            pnode = idx.internalPointer()
            self.modinfo = getattr(pnode, 'modinfo', None)
        self.accept()

    def dialog_cancel(self):
        self.reject()

@idlethreadsync
def selectStructNamespace(parent=None):
    d = VQStructNamespaceDialog(parent=parent)
    r = d.exec_()
    return d.modinfo




import struct

import envi
import envi.bits as e_bits
import envi.archs.i386 as e_i386
from . import opcode64 as opcode86

from envi.const import RMETA_NMASK

from envi.archs.i386.disasm import iflag_lookup, operand_range, priv_lookup, \
        i386Opcode, i386ImmOper, i386RegOper, i386ImmMemOper, i386RegMemOper, \
        i386SibOper, PREFIX_REPNZ, PREFIX_REPZ, PREFIX_REP, PREFIX_REP_SIMD, \
        PREFIX_OP_SIZE, PREFIX_ADDR_SIZE, MANDATORY_PREFIXES, PREFIX_REP_MASK,\
        RMETA_LOW8, RMETA_LOW16

from .regs import *
from envi.archs.i386.opconst import OP_EXTRA_MEMSIZES, OP_MEM_B, OP_MEM_W, OP_MEM_D, \
                                    OP_MEM_Q, OP_MEM_DQ, OP_MEM_QQ, OP_MEMMASK, \
                                    INS_VEXREQ, OP_NOVEXL, INS_VEXNOPREF, OP_NOREX, \
                                    PREFIX_REX, PREFIX_REX_B, PREFIX_REX_X, PREFIX_REX_W, \
                                    PREFIX_REX_MASK, PREFIX_REX_RXB, PREFIX_REX_R
all_tables = opcode86.tables86

# Pre generate these for fast lookup. Because our REX prefixes have the same relative
# bit relationship to eachother, we can cheat a little...
amd64_prefixes = list(e_i386.i386_prefixes)
amd64_prefixes[0x40] = (0x10 << 16)
amd64_prefixes[0x41] = (0x11 << 16)
amd64_prefixes[0x42] = (0x12 << 16)
amd64_prefixes[0x43] = (0x13 << 16)
amd64_prefixes[0x44] = (0x14 << 16)
amd64_prefixes[0x45] = (0x15 << 16)
amd64_prefixes[0x46] = (0x16 << 16)
amd64_prefixes[0x47] = (0x17 << 16)
amd64_prefixes[0x48] = (0x18 << 16)
amd64_prefixes[0x49] = (0x19 << 16)
amd64_prefixes[0x4a] = (0x1a << 16)
amd64_prefixes[0x4b] = (0x1b << 16)
amd64_prefixes[0x4c] = (0x1c << 16)
amd64_prefixes[0x4d] = (0x1d << 16)
amd64_prefixes[0x4e] = (0x1e << 16)
amd64_prefixes[0x4f] = (0x1f << 16)
amd64_prefixes[0xc5] = (0x20 << 16)  # VEX 2byte
amd64_prefixes[0xc4] = (0x40 << 16)  # VEX 3byte


REX_HIGH_DROP = ~(e_i386.RMETA_HIGH8 ^ e_i386.RMETA_LOW8)  # FIXME: this is ugly

PREFIX_VEX2  = 0x200000  # 2 byte VEX (data stored in opcode)
PREFIX_VEX3  = 0x400000  # 3 byte VEX (data stored in opcode)
PREFIX_VEX_L = 0x800000  # L bit set
PREFIX_VEX   = PREFIX_VEX2 | PREFIX_VEX3

PREFIX_SIZE_BOTH = (PREFIX_REX_W | PREFIX_VEX_L)

IMM_REQOFFS = (opcode86.ADDRMETH_I, opcode86.ADDRMETH_J, opcode86.ADDRMETH_L)

VEX_V_SHIFT  = 59

REX_BUMP = 8
MODE_16 = 0
MODE_32 = 1
MODE_64 = 2

RMETA_LOW32 = 0x200000

META_SIZES = [0 for i in range(9)]
META_SIZES[1] = RMETA_LOW8
META_SIZES[2] = RMETA_LOW16
META_SIZES[4] = RMETA_LOW32

class Amd64Opcode(i386Opcode):
    def __init__(self, va, opcode, mnem, prefixes, size, operands, iflags=0):
        '''
        Overriding this from envi/__init__.py in order to set the mnem for VEX instructions
        Technically this should be on the i386 one as well, but we don't yet support VEX for that.
        So oh well
        '''
        envi.Opcode.__init__(self, va, opcode, mnem, prefixes, size, operands, iflags)
        if prefixes & PREFIX_VEX and not opcode & INS_VEXNOPREF:
            mnem = 'v' + mnem
        self.mnem = mnem

    def __repr__(self):
        """
        Over-ride this if you want to make arch specific repr.
        """
        pfx = self.getPrefixName()
        if pfx:
            pfx = '%s: ' % pfx

        return pfx + self.mnem + " " + ",".join([o.repr(self) for o in self.opers])

    def render(self, mcanv):
        """
        Render this opcode to the specified memory canvas
        """
        if self.prefixes:
            pfx = self.getPrefixName()
            if pfx:
                mcanv.addNameText("%s: " % pfx, pfx)

        mcanv.addNameText(self.mnem, typename="mnemonic")
        mcanv.addText(" ")

        # Allow each of our operands to render
        imax = len(self.opers)
        lasti = imax - 1
        for i in range(imax):
            oper = self.opers[i]
            oper.render(mcanv, self, i)
            if i != lasti:
                mcanv.addText(",")

class Amd64RipRelOper(envi.DerefOper):
    def __init__(self, imm, tsize):
        self.imm = imm
        self.tsize = tsize
        self._is_deref = True

    def getOperValue(self, op, emu=None):
        if not self._is_deref:  # Special lea behavior
            return self.getOperAddr(op)
        if emu is None:
            return None
        return emu.readMemValue(self.getOperAddr(op, emu), self.tsize)

    def setOperValue(self, op, emu, val):
        emu.writeMemValue(self.getOperAddr(op, emu), val, self.tsize)

    def getOperAddr(self, op, emu=None):
        return op.va + op.size + self.imm

    def isDeref(self):
        # The disassembler may reach in and set this (if lea...)
        return self._is_deref

    def isDiscrete(self):
        return True

    def render(self, mcanv, op, idx):
        destva = op.va + op.size + self.imm
        sym = mcanv.syms.getSymByAddr(destva)

        mcanv.addNameText(e_i386.sizenames[self.tsize])
        mcanv.addText(" [")
        mcanv.addNameText("rip", typename="registers")

        if self.imm > 0:
            mcanv.addText(" + ")
            if sym is not None:
                mcanv.addVaText("$%s" % repr(sym), destva)
            else:
                mcanv.addNameText(str(self.imm))
        elif self.imm < 0:
            mcanv.addText(" - ")
            if sym is not None:
                mcanv.addVaText("$%s" % repr(sym), destva)
            else:
                mcanv.addNameText(str(abs(self.imm)))
        mcanv.addText("]")

    def repr(self, op):
        return "%s [rip + %d]" % (e_i386.sizenames[self.tsize], self.imm)


operands_index = 2
vex_pp_table = ( (0xf,), (0x66,), (0xf3,), (0xf2,) )
vex3_mmmm_table = ( (None,), (None,), (0x38,), (0x3A,), None, None, None, None )   # None is invalid.  (None,) adds no table depths


class Amd64Disasm(e_i386.i386Disasm):

    def __init__(self, mode=MODE_64):
        e_i386.i386Disasm.__init__(self, mode=mode)
        self._dis_oparch = envi.ARCH_AMD64
        self._dis_prefixes = amd64_prefixes
        self._dis_regctx = Amd64RegisterContext()
        self.ptrsize = 8

        # 64-bit only
        self._dis_amethods[opcode86.ADDRMETH_B >> 16] = self.ameth_b
        self._dis_amethods[opcode86.ADDRMETH_H >> 16] = self.ameth_h
        self._dis_amethods[opcode86.ADDRMETH_E >> 16] = self.ameth_e
        self._dis_amethods[opcode86.ADDRMETH_L >> 16] = self.ameth_l
        self._dis_amethods[opcode86.ADDRMETH_VEXH >> 16] = self.ameth_vexh

        # Over-ride these which are in use by the i386 version of the ASM
        self.ROFFSETSIMD  = e_i386.getRegOffset(amd64regs, "ymm0")
        self.ROFFSETDEBUG = e_i386.getRegOffset(amd64regs, "debug0")
        self.ROFFSETCTRL  = e_i386.getRegOffset(amd64regs, "ctrl0")
        self.ROFFSETTEST  = e_i386.getRegOffset(amd64regs, "test0")
        self.ROFFSETSEG   = e_i386.getRegOffset(amd64regs, "es")
        self.ROFFSETFPU   = e_i386.getRegOffset(amd64regs, "st0")
        # Note: getRegOffset doesn't work on meta registers and mm* are aliases of the
        # st registers, so we use getRegisterIndex instead
        self.ROFFSETMMX   = self._dis_regctx.getRegisterIndex("mm0")

    # NOTE: Technically, the REX must be the *last* prefix specified
    # NOTE: Technically, the VEX must be the *last* prefix specified (REX be damned)

    def _dis_calc_tsize(self, opertype, prefixes, operflags):
        """
        Use the oper type and prefixes to decide on the tsize for
        the operand.
        """
        mode = MODE_32

        sizelist = opcode86.OPERSIZE.get(opertype, None)
        if sizelist is None:
            raise Exception("OPERSIZE FAIL")

        if operflags & opcode86.OP_64AUTO:
            mode = MODE_64

        # NOTE: REX takes precedence over 66
        # (see section 2.2.1.2 in Intel 2a)
        if prefixes & PREFIX_SIZE_BOTH:

            mode = MODE_64

        elif prefixes & e_i386.PREFIX_OP_SIZE:

            mode = MODE_16

        return sizelist[mode]

    def disasm(self, bytez, offset, va, extra=None):
        '''
        The main amd64 decoder function. The inital steps it takes are determining what
        potential prefixes are attached to the instruction. By "potential", we mean that at
        this stage we don't know if thigs like 66, F2, F3 are being used as normal prefixes
        (representing things like a rep prefix) or if they're being used as mandatory prefixes
        that completely change with instruction we're decoding. All potential prefixes are stored
        in the pho_prefixes variable.

        To that end, there's some tap dancing we need to do to deal with what the intel manual
        refers to as "mandatory prefixes". If we hit a main opcode byte of 0F and we know we have a
        potentially mandatory prefix (and we're not in VEX land), we treat the byte right before the 0F
        as the only potential mandatory prefix (as laid out in the intel manual). Then we basically brute
        force the decoding since we really only have two paths to try. One where the mandatory prefix is
        merely a normal prefix (and doesn't affect which set of tables we traverse down) and one where
        the mandatory prefix does affect what tables we rachet through (and thus directly changes which
        instruction we're looking at). For each case, we append all the relevant output to a list (should
        the decoding produce a meaningful output).

        If we end up producing no instruction definitions from our brute force loop, we've hit an invalid
        sequence of instruction bytes and we throw an exception.

        If only one path produce output, then that's our results and we proceed on to use the instruction
        definition to determine what addressing methods and size types to use when determining operands.

        If both paths produce a valid instruction definition, then the path that uses the mandatory prefix
        to directly change the instruction takes precedence over the path where it's just a normal prefix.

        In both the one and two results case, outside of our instruction decoding loop, we've kept a list of
        the possible decodings we could have hit, and just merely pop off the end of the list (so order
        matters when building the ppref variable).
        '''
        # FIXME: for newer instructions, the VEX.W bit needs to be able to change the opcode. ugh.
        # FIXME: And also REX.W

        # Stuff for opcode parsing
        tabdesc = all_tables[opcode86.TBL_Main]  # A tuple (optable, shiftbits, mask byte, sub, max)
        startoff = offset  # Use startoff as a size knob if needed
        isvex = False
        vexw = None
        last_pref = 0
        ppref = [(None, None)]

        # Stuff we'll be putting in the opcode object
        optype = None  # This gets set if we successfully decode below
        mnem = None
        operands = []

        prefixes = 0
        pho_prefixes = 0  # faux prefixes...don't immediately apply them, they may not be the prefixes we're looking for

        while True:

            obyte = bytez[offset]

            # This line changes in 64 bit mode
            p = self._dis_prefixes[obyte]
            if p is None:
                break

            if MANDATORY_PREFIXES[obyte]:
                pho_prefixes |= p
                last_pref = obyte
            else:
                prefixes |= p

            if p & PREFIX_VEX:
                isvex = True
                if p == PREFIX_VEX2:
                    offset += 1
                    imm1 = bytez[offset]
                    # shouldn't in 64-bit mode, but in 32-bit, this keeps LES from colliding
                    # TODO: So we're always in 64 bit here. This will need to be here once we unify 32/64 decoding
                    #if imm1 & 0xc0 != 0xc0:
                        #break
                    inv1 = imm1 ^ 0xff

                    vex_l = (0, PREFIX_VEX_L)[(imm1 & 4) >> 2]
                    vvvv = ((inv1 >> 3) & 0xf)
                    pp = imm1 & 3

                    prefixes |= (inv1 << 11) & PREFIX_REX_R     # R is inverted
                    prefixes |= vex_l
                    prefixes |= (vvvv << VEX_V_SHIFT)
                    combined_mand_prefixes = vex_pp_table[pp]

                elif p == PREFIX_VEX3:
                    imm1 = bytez[offset+1]
                    offset += 2
                    # TODO: So we're always in 64 bit here. This will need to be here once we unify 32/64 decoding
                    #if imm1 & 0xc0 != 0xc0:
                        #break
                    imm2 = bytez[offset]
                    inv1 = imm1 ^ 0xff
                    inv2 = imm2 ^ 0xff

                    vex_l = (0, PREFIX_VEX_L)[(imm2 & 4) >> 2]
                    vvvv = ((inv2 >> 3) & 0xf)
                    pp = imm2 & 3
                    m_mmmm = imm1 & 0x1f
                    prefixes |= ((inv1 << 11) & PREFIX_REX_RXB)     # RXB are inverted
                    vexw = ((imm2 << 12) & PREFIX_REX_W)            # W is not inverted
                    prefixes |= vexw
                    prefixes |= vex_l
                    prefixes |= (vvvv << VEX_V_SHIFT)               # vvvv

                    combined_mand_prefixes = vex_pp_table[ pp ] + vex3_mmmm_table[m_mmmm]

                # VEX prefixes default to 0F table, possibly F20F, F30F or 660F
                # VEX3 prefixes may also specify depths into 38 and 3A tables
                for tabidx in combined_mand_prefixes:
                    if tabidx is None:
                        continue
                    opdesc = tabdesc[0][tabidx]
                    tabdesc = all_tables[opdesc[0]]
                # So VEX and mandatory prefixes don't really intermingle
                offset += 1
                break

            offset += 1
            continue

        if obyte != 0x0f:
            prefixes |= pho_prefixes

        # intel manual says VEX and legacy prefixes don't intermingle
        if obyte == 0x0f and MANDATORY_PREFIXES[last_pref] and not isvex:
            obyte = last_pref
            ppref.append((last_pref, amd64_prefixes[last_pref]))

        decodings = []
        mainbyte = offset
        all_prefixes = prefixes

        ogtabdesc = tabdesc
        # onehot in this case refers to the their prefixes that are defined in i386/disasm.py where only
        # on bit of the entire integer is set. We use that to quickly pop things in and out of the prefixes
        # list
        for pref, onehot in ppref:
            tabdesc = ogtabdesc
            offset = mainbyte
            if pref is not None:
                # our mandatory prefix is not none, which means that we have to jump through the tables
                # using the mandatory prefix byte as our "main byte"
                # As a bit of a hack, the 66/F2/F3 entries in the main table
                # directly point to the 660F/F20F/F30F tables since we're carefully tap dancing around
                # what our opcode byte really is
                obyte = pref
                # since we're treating this prefix as mandatory and not as REPNZ/REPZ/etc, we need to rip
                # it out of the pho_prefixes before we combine pho_prefixes with the main prefixes container
                all_prefixes = prefixes | (pho_prefixes & (~onehot))
            else:
                # treat nothing as a mandatory prefix (or we defaulted into here if we got no mandatory
                # prefixes). For most instructions this will be the normal case.
                obyte = bytez[offset]
                all_prefixes = prefixes | pho_prefixes

            while True:
                if (obyte > tabdesc[5]):
                    tabdesc = all_tables[tabdesc[6]]

                tabidx = ((obyte - tabdesc[4]) >> tabdesc[2]) & tabdesc[3]
                opdesc = tabdesc[0][tabidx]

                # Hunt down multi-byte opcodes
                nexttable = opdesc[0]
                if nexttable != 0:  # If we have a sub-table specified, use it.
                    tabdesc = all_tables[nexttable]

                    # Account for the table jump we made
                    offset += 1
                    obyte = bytez[offset]
                    continue

                # We are now on the final table...
                tbl_opercnt = tabdesc[1]
                mnem = opdesc[3 + tbl_opercnt]
                optype = opdesc[1]
                if tabdesc[3] == 0xff:
                    offset += 1  # For our final opcode byte
                break

            if optype & INS_VEXREQ and not isvex:
                continue

            if optype != 0:
                decodings.append((tabdesc, opdesc, offset, all_prefixes))

        if not len(decodings):
            raise envi.InvalidInstruction(bytez=bytez[startoff:startoff+16], va=va)

        tabdesc, opdesc, offset, prefixes = decodings.pop()
        optype = opdesc[1]
        tbl_opercnt = tabdesc[1]
        mnem = opdesc[3 + tbl_opercnt]

        if optype == 0:
            raise envi.InvalidInstruction(bytez=bytez[startoff:startoff+16], va=va)

        operoffset = 0
        # Begin parsing operands based off address method
        for i in range(operands_index, operands_index + tbl_opercnt):

            oper = None  # Set this if we end up with an operand
            osize = 0

            # Pull out the operand description from the table
            operflags = opdesc[i]
            opertype = operflags & opcode86.OPTYPE_MASK
            addrmeth = operflags & opcode86.ADDRMETH_MASK

            # If there are no more operands, break out of the loop!
            if operflags == 0:
                break

            # handles tsize calculations including new REX prefixes
            tsize = self._dis_calc_tsize(opertype, prefixes, operflags)
            # If addrmeth is zero, we have operands embedded in the opcode
            if addrmeth == 0:
                osize = 0
                oper = self.ameth_0(operflags, opdesc[2+tbl_opercnt+i], tsize, prefixes)

            else:
                # So the 0x7f is here to help us deal with an issue between VEX and non-VEX
                # A super common patter in vex is to add an operand somewhere in the middle of the
                # existing operands. So if we have like cmpps xmm2, 17 in non-VEX, the vex version
                # will look like vsprlw xmm3, xmm4, 17.
                # The fun bit of this is that the vex only portions aren't exclusive to the VEX-only
                # addressing methods, so we can have ADDRMETH_V be skipped outside of VEX mode too, and not
                # just things like ADDRMETH_H. Hence, we need a new flag that I stash in the upper bits of
                # instruction operand definition so we can know when to skip operands
                ameth = self._dis_amethods[(addrmeth >> 16) & 0x7F]
                vex_skip = addrmeth & opcode86.ADDRMETH_VEXSKIP
                if not isvex and vex_skip:
                    continue

                if ameth is None:
                    raise Exception("Implement Addressing Method 0x%.8x" % addrmeth)

                # NOTE: Depending on your addrmethod you may get beginning of operands, or offset
                try:
                    if addrmeth in IMM_REQOFFS:
                        osize, oper = ameth(bytez, offset+operoffset, tsize, prefixes, operflags)

                        # If we are a sign extended immediate and not the same as the other operand,
                        # do the sign extension during disassembly so nothing else has to worry about it..
                        if operflags & opcode86.OP_SIGNED:
                            if len(operands) and tsize != operands[-1].tsize:
                                otsize = operands[-1].tsize
                                oper.imm = e_bits.sign_extend(oper.imm, oper.tsize, otsize)
                                oper.tsize = otsize
                            elif not len(operands):
                                oper.imm = e_bits.sign_extend(oper.imm, oper.tsize, self._dis_default_size)
                                oper.tsize = self._dis_default_size

                    else:
                        # see same code section in i386 for this rationale
                        osize, oper = ameth(bytez, offset, tsize, prefixes, operflags)
                        if oper and oper.isDeref():
                            memsz = OP_EXTRA_MEMSIZES[(operflags & OP_MEMMASK) >> 4]
                            if memsz is not None:
                                oper.tsize = memsz
                            if prefixes & PREFIX_ADDR_SIZE:
                                if getattr(oper, 'reg', None) is not None:
                                    oper.reg |= RMETA_LOW32
                                elif getattr(oper, 'index', None) is not None:
                                    oper.index |= RMETA_LOW32

                except struct.error:
                    # Catch struct unpack errors due to insufficient data length
                    raise envi.InvalidInstruction(bytez=bytez[startoff:startoff+16])

            if oper is not None:
                # This is a filty hack for now...
                oper._dis_regctx = self._dis_regctx
                operands.append(oper)

            operoffset += osize

        typemask = optype & 0xFFFF
        # Pull in the envi generic instruction flags
        iflags = iflag_lookup.get(typemask, 0) | self._dis_oparch

        if prefixes & PREFIX_REP_MASK:
            iflags |= envi.IF_REPEAT

        if priv_lookup.get(mnem, False):
            iflags |= envi.IF_PRIV

        # Lea will have a reg-mem/sib operand with _is_deref True, but should be false
        if typemask == opcode86.INS_LEA:
            operands[1]._is_deref = False

        ret = Amd64Opcode(va, optype, mnem, prefixes, (offset-startoff)+operoffset, operands, iflags)
        return ret

    def parse_modrm(self, byte, prefixes=0):
        # Pass in a string with an offset for speed rather than a new string
        mod = (byte >> 6) & 0x3
        reg = (byte >> 3) & 0x7
        rm = byte & 0x7

        if prefixes & PREFIX_REX_R:
            reg |= 0b1000

        if prefixes & PREFIX_REX_B:
            if not (mod != 3 and rm == 4):  # if not SIB
                rm |= 0b1000

        return (mod,reg,rm)

    def byteRegOffset(self, val, prefixes=0):
        # NOTE: Override this because there is no AH etc in 64 bit mode
        if (prefixes & PREFIX_REX):     # the parse_modrm function deals with register index adds
            val |= e_i386.RMETA_LOW8

        else:  # not using REX, revert to old split-registers (low/high)
            if val < 4:
                val |= e_i386.RMETA_LOW8
            else:
                val |= e_i386.RMETA_HIGH8
                val -= 4

        return val

    def extended_parse_modrm(self, bytes, offset, opersize, regbase=0, prefixes=0):
        """
        Return a tuple of (size, Operand)
        """
        size = 1
        # FIXME this would be best to not parse_modrm twice.  tweak it.
        mod,reg,rm = self.parse_modrm(bytes[offset], prefixes)
        if mod == 0 and rm == 5:
            imm = e_bits.parsebytes(bytes, offset + size, 4, sign=True)
            size += 4
            return(size, Amd64RipRelOper(imm, opersize))

        return e_i386.i386Disasm.extended_parse_modrm(self, bytes, offset, opersize, regbase, prefixes)

    # NOTE: Override a bunch of the address modes to account for REX
    def ameth_0(self, operflags, operval, tsize, prefixes):
        # o = e_i386.i386Disasm.ameth_0(self, operflags, operval, tsize, prefixes)
        if operflags & opcode86.OP_REG:
            # for handling meta registers embedded in opcodes
            if prefixes & PREFIX_OP_SIZE:
                if self._dis_regctx.isMetaRegister(operval):
                    operval = (operval & RMETA_NMASK) | META_SIZES[tsize]
                else:
                    operval |= META_SIZES[tsize]
            width = self._dis_regctx.getRegisterWidth(operval) >> 3
            o = i386RegOper(operval, width)
        elif operflags & opcode86.OP_IMM:
            o = i386ImmOper(operval, tsize)
        else:
            raise Exception("Unknown ameth_0! operflags: 0x%.8x" % operflags)

        if not (operflags & OP_NOREX) and prefixes & PREFIX_REX_W and getattr(o, 'reg', None):
            o.reg &= 0xffff

        if prefixes & PREFIX_REX_B and isinstance(o, e_i386.i386RegOper):
            # the optable entries for AH with REX_B is terribly unhelpful.
            if o.reg & e_i386.RMETA_HIGH8 == e_i386.RMETA_HIGH8:
                o.reg &= REX_HIGH_DROP
                o.reg += 4
            if not (operflags & OP_NOREX):
                o.reg += REX_BUMP
        return o

    def ameth_vexh(self, bytes, offset, tsize, prefixes, operflags):
        '''
        So this is here because instructions like movss and movsd are ambiguous in their
        2/3 operand state without first jumping ahead to the modrm byte. If modrm refers to
        memory, we skip this state and just go on to the next addressing method. If it refers
        to a register, pass through to self.ameth_h
        '''
        mod, reg, rm = self.parse_modrm(bytes[offset], prefixes)
        if mod == 3:
            return self.ameth_h(bytes, offset, tsize, prefixes, operflags)
        else:
            return (0, None)

    def ameth_l(self, bytez, offset, tsize, prefixes, operflags):
        reg = self.ROFFSETSIMD
        imm = e_bits.parsebytes(bytez, offset, 1, sign=False)
        idx = (imm & 0xF0) >> 4
        if not (prefixes & PREFIX_VEX_L) or operflags & OP_NOVEXL:
            reg += e_i386.RMETA_LOW128
        return (1, i386RegOper(reg + idx, tsize))

    def ameth_g(self, bytes, offset, tsize, prefixes, operflags):
        osize, oper = e_i386.i386Disasm.ameth_g(self, bytes, offset, tsize, prefixes, operflags)
        # TODO: Disallowing reg_rip is probably wrong
        # TODO: the addr override operates off the default of the instruction, so we need to grab that
        if oper.tsize == 4:
            oper.reg |= RMETA_LOW32
        if prefixes & PREFIX_REX_R:
            oper.reg += REX_BUMP
        return osize, oper

    def ameth_b(self, bytez, offset, tsize, prefixes, operflags):
        osize = 0
        oper = 0
        vvvv = (prefixes >> VEX_V_SHIFT) & 0xf
        oper = i386RegOper(vvvv, tsize)
        # TODO: Disallowing reg_rip is probably wrong
        if oper.tsize == 4:
            oper.reg |= RMETA_LOW32
        return osize, oper

    def ameth_h(self, bytez, offset, tsize, prefixes, operflags):
        osize = 0
        vvvv = (prefixes >> VEX_V_SHIFT) & 0xf
        offset = self.ROFFSETSIMD
        if not (prefixes & PREFIX_VEX_L) or operflags & OP_NOVEXL:
            vvvv |= e_i386.RMETA_LOW128

        oper = i386RegOper(offset + vvvv, tsize)
        return osize, oper

    def ameth_c(self, bytes, offset, tsize, prefixes, operflags):
        osize, oper = e_i386.i386Disasm.ameth_c(self, bytes, offset, tsize, prefixes, operflags)
        if prefixes & PREFIX_REX_R:
            oper.reg += REX_BUMP
        return osize, oper

    def ameth_d(self, bytes, offset, tsize, prefixes, operflags):
        osize, oper = e_i386.i386Disasm.ameth_d(self, bytes, offset, tsize, prefixes, operflags)
        if prefixes & PREFIX_REX_R:
            oper.reg += REX_BUMP
        return osize, oper

    def ameth_v(self, bytes, offset, tsize, prefixes, operflags):
        osize, oper = e_i386.i386Disasm.ameth_v(self, bytes, offset, tsize, prefixes, operflags)

        # FIXME: YMM->XMM (this may move into i386 version when VEX is made available there)
        if not (prefixes & PREFIX_VEX_L) or operflags & OP_NOVEXL:
            oper.reg += e_i386.RMETA_LOW128

        if prefixes & PREFIX_REX_R:
            oper.reg += REX_BUMP
        return osize, oper

    def ameth_u(self, bytes, offset, tsize, prefixes, operflags):
        osize, oper = e_i386.i386Disasm.ameth_u(self, bytes, offset, tsize, prefixes, operflags)

        if not (prefixes & PREFIX_VEX_L) or operflags & OP_NOVEXL:
            oper.reg += e_i386.RMETA_LOW128

        if prefixes & PREFIX_REX_B:
            oper.reg += REX_BUMP

        return osize, oper

    # NOTE: The ones below are the only ones to which REX.X or REX.B can apply (besides ameth_0)
    # FIXME: we need to adhere to the ADDR_SIZE/OPER_SIZE rules better...  the REX_BUMP is an afterthought...
    # however, all the rules are based on ADDR_SIZE/OPER_SIZE and can be coalesced elegantly into amd64/i386
    # and even make 16-bit mode play nicely.
    # FIXME: processing of REX into actual operand size/indexes should be done *pre* operand-parsing
    # if for no other reason, because that's how the manuals discuss it.
    # NO REALLY ! FIXME!
    def _dis_rex_exmodrm(self, oper, prefixes, operflags):
        # REMEMBER: all extended mod RM reg fields come from the r/m part.  If it
        #           were actually just the reg part, it'd be in one of the above
        #           addressing modes...

        if hasattr(oper, "index"):
            # Adjust the size if needed
            if prefixes & PREFIX_REX_X:
                # fix up index in case it was set to None by sib_parse (index==4 without REX)
                if oper.index is None:
                    oper.index = 4
                oper.index += REX_BUMP

        # oper.reg will be r/m or SIB base
        if getattr(oper, "reg", None) is not None:
            # Adjust the size if needed
            if prefixes & PREFIX_REX_B:
                oper.reg += REX_BUMP

        if isinstance(oper, e_i386.i386RegOper):
            if oper.tsize == 4:
                oper.reg |= RMETA_LOW32

    def ameth_e(self, bytes, offset, tsize, prefixes, operflags):
        regbase = 0
        # TODO: Does this impact memory as well?
        osize, oper = self.extended_parse_modrm(bytes, offset, tsize, prefixes=prefixes, regbase=regbase)
        self._dis_rex_exmodrm(oper, prefixes, operflags)
        return osize, oper

    def ameth_w(self, bytez, offset, tsize, prefixes, operflags):
        mod, reg, rm = self.parse_modrm(bytez[offset])
        if mod == 3:
            vvvv = self.ROFFSETSIMD
            if not (prefixes & PREFIX_VEX_L) or operflags & OP_NOVEXL:
                vvvv |= e_i386.RMETA_LOW128
                tsize = 16

            osize, oper = (1, i386RegOper(rm + vvvv, tsize))
        else:
            osize, oper = self.extended_parse_modrm(bytez, offset, tsize, prefixes=prefixes)
            if (oper.tsize == 32 and not (prefixes & PREFIX_VEX_L)) or operflags & OP_NOVEXL:
                oper.tsize = 16

        self._dis_rex_exmodrm(oper, prefixes, operflags)
        return osize, oper


if __name__ == '__main__':
    import envi.archs
    envi.archs.dismain(Amd64Disasm())



import sys
import struct
import traceback

import vdb.testmods
import vdb.testmods.hookbptest as v_t_hookbptest
import vdb.testmods.callingconventions as v_t_callingconventions

tests = [
]

tests32 = [
    v_t_callingconventions.i386StdCallCallingConventionTest(0),
    v_t_callingconventions.i386CdeclCallingConventionTest(0),
    v_t_callingconventions.i386ThisCallCallingConventionTest(0),
    v_t_callingconventions.i386MsFastCallCallingConventionTest(0),
    v_t_callingconventions.i386BFastCallCallingConventionTest(0),

    v_t_callingconventions.i386StdCallCallingConventionTest(1),
    v_t_callingconventions.i386CdeclCallingConventionTest(1),
    v_t_callingconventions.i386ThisCallCallingConventionTest(1),
    v_t_callingconventions.i386MsFastCallCallingConventionTest(1),
    v_t_callingconventions.i386BFastCallCallingConventionTest(1),

    v_t_callingconventions.i386StdCallCallingConventionTest(3),
    v_t_callingconventions.i386CdeclCallingConventionTest(3),
    v_t_callingconventions.i386ThisCallCallingConventionTest(3),
    v_t_callingconventions.i386MsFastCallCallingConventionTest(3),
    v_t_callingconventions.i386BFastCallCallingConventionTest(3),

    v_t_callingconventions.i386StdCallCallingConventionTest(10),
    v_t_callingconventions.i386CdeclCallingConventionTest(10),
    v_t_callingconventions.i386ThisCallCallingConventionTest(10),
    v_t_callingconventions.i386MsFastCallCallingConventionTest(10),
    v_t_callingconventions.i386BFastCallCallingConventionTest(10),
]

tests64 = [
    v_t_callingconventions.x64MSx64CallCallingConventionTest(0),
    v_t_callingconventions.x64MSx64CallCallingConventionTest(1),
    v_t_callingconventions.x64MSx64CallCallingConventionTest(3),
    v_t_callingconventions.x64MSx64CallCallingConventionTest(10),

    v_t_callingconventions.SysVAmd64CallCallingConventionTest(0),
    v_t_callingconventions.SysVAmd64CallCallingConventionTest(1),
    v_t_callingconventions.SysVAmd64CallCallingConventionTest(3),
    v_t_callingconventions.SysVAmd64CallCallingConventionTest(10),

    v_t_callingconventions.SysVAmd64SystemCallCallingConventionTest(0),
    v_t_callingconventions.SysVAmd64SystemCallCallingConventionTest(1),
    v_t_callingconventions.SysVAmd64SystemCallCallingConventionTest(3),
    v_t_callingconventions.SysVAmd64SystemCallCallingConventionTest(10),
]

windows = [
    v_t_hookbptest.HookBpTest(),
    v_t_hookbptest.HookBpTest2(),
    v_t_hookbptest.HookBpTest3(),
    v_t_hookbptest.HookBpTest4(),
    v_t_hookbptest.HookBpTest5(),
    v_t_hookbptest.HookBpTest6(),
]

def runTests(tests):

    for test in tests:
        testname = test.__class__.__name__
        try:
            stage = 'prep'
            test.prepTest()
            stage = 'run'
            test.runTest()
            stage = 'clean'
            test.cleanTest()
            print('Test Success: %s' % test.__class__.__name__)
        except vdb.testmods.SkipTest as e:
            print('Test Skipped: %s (in %s) %s' % (testname, stage, e))
        except Exception as e:
            traceback.print_exc()
            print('Test Failure: %s (in %s) %s' % (testname, stage, e))

def main():

    print('running bare minimum vtrace tests')
    runTests(tests)
    print('')
    print('running architecture specific tests')
    if struct.calcsize('P') == 4:
        runTests(tests32)
    elif struct.calcsize('P') == 8:
        runTests(tests64)
    else:
        raise Exception('unsupported bitness')
    print('')

    print('running platform specific tests')
    if 'win' in sys.platform.lower():
        runTests(windows)


if __name__ == '__main__':
    sys.exit(main())



import os

import vtrace
import vtrace.tests as vt_tests

class ThreadNotifier(vtrace.Notifier):

    def __init__(self):
        vtrace.Notifier.__init__(self)
        self.threadexit = False
        self.threadcreate = False

    def notify(self, event, trace):
        if event == vtrace.NOTIFY_CREATE_THREAD:
            self.threadcreate = True
            return

        if event == vtrace.NOTIFY_EXIT_THREAD:
            self.threadexit = True
            return

class VtraceThreadTest(vt_tests.VtraceProcessTest):

    pypath = os.path.join('vtrace','tests','mains','mainthreads.py')

    def test_vtrace_threads(self):
        n = ThreadNotifier()

        self.trace.registerNotifier(vtrace.NOTIFY_ALL, n)
        self.runUntilExit()

        self.assertTrue(n.threadcreate)
        self.assertTrue(n.threadexit)



import time
import unittest
import itertools
import threading

import cobra
import cobra.auth as c_auth
import cobra.auth.shadowfile as c_auth_shadow

import cobra.tests as c_tests

class CobraBasicTest(unittest.TestCase):

    def test_cobra_proxy(self):

        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=60600)
        objname = daemon.shareObject( testobj )
        daemon.fireThread()

        t = cobra.CobraProxy('cobra://localhost:60600/%s' % objname)
        c_tests.accessTestObject( t )

        daemon.stopServer()

    def test_cobra_msgpack(self):

        try:
            import msgpack
        except ImportError:
            self.skipTest('No msgpack installed!')

        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=60610, msgpack=True)
        objname = daemon.shareObject( testobj )
        daemon.fireThread()

        t = cobra.CobraProxy('cobra://localhost:60610/%s?msgpack=1' % objname)
        c_tests.accessTestObject( t )
        daemon.stopServer()

    def test_cobra_authentication(self):

        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=60601)
        daemon.setAuthModule( c_auth.CobraAuthenticator() )
        daemon.fireThread()

        objname = daemon.shareObject( testobj )

        # Lets fail because of no-auth first
        try:
            p = cobra.CobraProxy('cobra://localhost:60601/%s' % objname)
            raise Exception('Allowed un-authd connection!')
        except cobra.CobraAuthException as e:
            pass

        # Now fail with wrong auth
        try:
            p = cobra.CobraProxy('cobra://localhost:60601/%s' % objname, authinfo={})
            raise Exception('Allowed bad-auth connection!')
        except cobra.CobraAuthException as e:
            pass

        # Now lets succeed
        authinfo = { 'user':'invisigoth', 'passwd':'secret' }
        t = cobra.CobraProxy('cobra://localhost:60601/%s' % objname, authinfo=authinfo)
        c_tests.accessTestObject( t )

        daemon.stopServer()

    def test_cobra_shadowauth(self):
        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=60602)
        shadowfile = c_tests.testFileName('shadowpass.txt')
        authmod = c_auth_shadow.ShadowFileAuth( shadowfile )
        daemon.setAuthModule( authmod )

        daemon.fireThread()

        objname = daemon.shareObject( testobj )

        # Now lets succeed
        authinfo = { 'user':'invisigoth', 'passwd':'secret' }
        t = cobra.CobraProxy('cobra://localhost:60602/%s' % objname, authinfo=authinfo)
        c_tests.accessTestObject(t)
        self.assertEqual( t.getUser(), 'invisigoth')
        daemon.stopServer()

    def test_cobra_refcount(self):

        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=60660)
        objname = daemon.shareObject( testobj, doref=True )
        daemon.fireThread()

        with cobra.CobraProxy('cobra://localhost:60660/%s' % objname) as t:
            c_tests.accessTestObject( t )

        self.assertIsNone( daemon.getSharedObject( objname ) )
        daemon.stopServer()

    #def test_cobra_ssl(self):
    #def test_cobra_ssl_clientcert(self):
    def test_cobra_helpers(self):
        portnum = itertools.count(60651)

        testobj = c_tests.TestObject()

        # test startCobraServer
        port = None
        daemon = None
        while daemon is None:
            try:
                port = next(portnum)
                daemon = cobra.startCobraServer(host="", port=port, sslca=None, sslcrt=None, sslkey=None, msgpack=True)

            except Exception as e:
                logger.warning("exception starting cobra server: %r", e)

        objname = daemon.shareObject( testobj )
        tproxy = cobra.CobraProxy('cobra://localhost:%d/%s?msgpack=1' % (port, objname))
        c_tests.accessTestObject( tproxy )

        # test getCobraDaemon()
        port = next(portnum)
        testobj = c_tests.TestObject()

        daemon = cobra.CobraDaemon(port=port, msgpack=True)
        cobra.daemon_manager.registerCobraDaemon(daemon)
        daemon.fireThread()
        d2 = cobra.getCobraDaemon(host="", port=port, sslca=None, sslcrt=None, sslkey=None, msgpack=True, create=False)
        objname = d2.shareObject( testobj )
        tproxy = cobra.CobraProxy('cobra://localhost:%d/%s?msgpack=1' % (port, objname))
        c_tests.accessTestObject( tproxy )

        cobra.daemon_manager.deregisterCobraDaemon('', port)

        port = next(portnum)
        testobj = c_tests.TestObject()

        daemon = cobra.getCobraDaemon(host="", port=port, sslca=None, sslcrt=None, sslkey=None, msgpack=True)
        objname = daemon.shareObject( testobj )
        tproxy = cobra.CobraProxy('cobra://localhost:%d/%s?msgpack=1' % (port, objname))
        c_tests.accessTestObject( tproxy )

if __name__ == '__main__':
    unittest.main()




'''
A module to contain code flow analysis for envi opcode objects...
'''
import logging
import collections

import envi
import envi.common as e_cmn
import envi.memory as e_mem
import envi.const as e_const

logger = logging.getLogger(__name__)


class CodeFlowContext(object):

    '''
    A CodeFlowContext is used for code-flow (not linear) based disassembly
    for an envi MemoryObject (which is responsible for knowing the
    implementation of parseOpcode().  The CodeFlowContext will optionally
    notify several callback handlers for different events which occur during
    disassembly:

    self._cb_opcode(va, op, branches) - called for every newly parsed opcode
        NOTE: _cb_opcode must return the desired branches for continued flow

    self._cb_function(fva, metadict) - called once for every function

    self._cb_branchtable(tabva, ptrva, destva) - called for switch tables
        NOTE: Return False to stop iteration of pointers

    Set exptable=True to expand branch tables in this phase
    Set persist=True to never disasm the same thing twice
    Set recurse=True to automatically code flow to nested functions
    '''
    def __init__(self, mem, persist=False, exptable=True, recurse=True):

        self._funcs = {}
        self._fcalls = {}
        self._mem = mem
        self._cf_noret = {}     # noret funcs
        self._cf_noflow = {}    # va's to stop on

        # A few options to the codeflow object
        self._cf_persist = None
        if persist:
            self._cf_persist = {}

        self._cf_recurse = recurse
        self._cf_exptable = exptable
        self._cf_blocks = []

        self._cf_blocked = collections.OrderedDict()
        self._cf_delaying = collections.defaultdict(set)
        self._cf_delayed = collections.defaultdict(set)
        self._calls_from = {}

        self._dynamic_branch_handlers = []

    def _cb_opcode(self, va, op, branches):
        '''
        Extend CodeFlowContext and implement this method to receive
        a callback for every newly discovered opcode.
        '''
        return branches

    def _cb_function(self, fva, fmeta):
        '''
        Extend CodeFlowContext and implement this method to recieve
        a callback for every newly discovered function.  Additionally,
        metadata about the function may be stored in the fmeta dict.
        '''
        pass

    def _cb_noflow(self, va, tva):
        '''
        Implement this method to receive a callback when a given code
        branch is skipped due to being in the noflow dictionary.
        ( likely due to procedural branch to noreturn address )
        '''
        pass

    def _cb_branchtable(self, tableva, ptrva, destva):
        '''
        Extend CodeFlowContext and implement this method to receive
        a callback for every conditional branch in a discovered
        "branch table" ( think jump/switch cases ).
        tableva     - The base address of the table
        ptrva       - The address of the pointer for this index
        destva      - The destination address (deref of ptrva)

        Return False to stop table iteration.
        '''
        pass

    def _cb_dynamic_branch(self, va, op, bflags, branches):
        '''
        if codeflow finds a branch to a non-discrete value (eg: to a register)
        we handle it here. By default, we simply track the dynamic branch in a global
        VaSet which is added to every workspace.
        '''
        '''
        When code-flow analysis runs into an indirect branch it doesn't know
        what to do with, the architecture can take a crack at it.
        '''
        for cb in self._dynamic_branch_handlers:
            cb(self, op, self._mem, bflags, branches)

    def addNoReturnAddr(self, va):
        '''
        Add a virtual address to the list of VAs that are non-returning
        procedural branch targets.
        '''
        self._cf_noret[va] = True

    def addNoFlow(self, va, destva):
        '''
        Add a va->destva no-flow entry which will prevent codeflow from
        continuing to destva as a result of va ( destva may still be
        decoded as a result of being reached some other way... )
        '''
        logger.debug("addNoFlow(0x%x, 0x%x)", va, destva)
        self._cf_noflow[(va, destva)] = True

    def getCallsFrom(self, fva):
        return self._fcalls.get(fva)

    def addFunctionDef(self, fva, calls_from):
        '''
        Add a priori knowledge of a function to the code flow
        stuff...
        '''
        self._fcalls[fva] = calls_from

    def addCodeFlow(self, va, arch=envi.ARCH_DEFAULT):
        '''
        Do code flow disassembly from the specified address. Returns a list
        of the procedural branch targets discovered during code flow...

        Set persist=True to store 'opdone' and never disassemble the same thing twice
        '''
        logger.debug('addCodeFlow(0x%x, 0x%x)', va, arch)
        opdone = {}
        if self._cf_persist is not None:
            opdone = self._cf_persist

        calls_from = {}
        optodo = [((0, va), arch), ]
        startva = va
        self._cf_blocks.append(va)
        cf_eps = collections.OrderedDict()
        while len(optodo):

            todo, arch = optodo.pop()

            if self._cf_noflow.get(todo):
                self._cb_noflow(*todo)
                continue

            pva, va = todo

            if opdone.get(va):
                continue

            opdone[va] = True

            try:
                op = self._mem.parseOpcode(va, arch=arch)
            except envi.InvalidInstruction as e:
                logger.warning('parseOpcode error at 0x%.8x (addCodeFlow(0x%x)): %s', va, startva, e)
                continue
            except Exception as e:
                logger.warning('Codeflow exception at 0x%.8x (addCodeFlow(0x%x)): %s', va, startva, e)
                continue

            branches = op.getBranches()
            # The opcode callback may filter branches...
            branches = self._cb_opcode(va, op, branches)

            # FIXME: if IF_BRANCH: if IF_COND and len(branches)<2: _cb_dynamic_branch()
            # FIXME: if IF_BRANCH and not IF_COND and len(branches)<1: _cb_dynamic_branch()
            # FIXME: if IF_CALL and len(branches)<2: _cb_dynamic_branch()

            while len(branches):

                bva, bflags = branches.pop()

                # look for dynamic branches (ie. branches which don't have a known target).
                # Assume at least one branch
                if bva is None:
                    self._cb_dynamic_branch(va, op, bflags, branches)

                if self._cf_noflow.get((va, bva)):
                    self._cb_noflow(va, bva)
                    continue

                try:
                    # Handle a table branch by adding more branches...
                    # most, if not all, of the work to construct jump tables is done in makeOpcode
                    if bflags & envi.BR_TABLE:
                        if self._cf_exptable:
                            ptrbase = bva
                            tabdone = set()
                            for bdest in self._mem.iterJumpTable(ptrbase):
                                if not self._cb_branchtable(bva, ptrbase, bdest):
                                    break
                                if bdest not in tabdone:
                                    tabdone.add(bdest)
                                    branches.append((bdest, envi.BR_COND))
                                ptrbase += self._mem.psize
                        continue

                    if bflags & envi.BR_DEREF:

                        if not self._mem.probeMemory(bva, self._mem.psize, e_const.MM_READ):
                            continue

                        # Before we update bva, lets check if its in noret...
                        if self._cf_noret.get(bva):
                            self.addNoFlow(va, va + len(op))

                        bva = self._mem.readMemoryPtr(bva)

                    if not self._mem.probeMemory(bva, 1, e_const.MM_EXEC):
                        continue

                    if self._mem.probeMemory(bva, 1, e_const.MM_UNINIT):
                        continue

                    if bflags & envi.BR_PROC:

                        # Record that the current code flow has a call from it
                        # to the branch target...
                        nextva = va + len(op)

                        if bva != nextva:  # NOTE: avoid call 0 constructs

                            # Now we descend so we do deepest func callbacks first!
                            if self._cf_recurse:
                                # descend into functions, but make sure we don't descend into
                                # recursive functions
                                if bva in self._cf_blocks:
                                    logger.debug("not recursing to function 0x%x (at 0x%x): it's already in analysis call path (ie. it called *this* func)",
                                            bva, va)
                                    logger.debug("call path: \t" + ", ".join([hex(x) for x in self._cf_blocks]))
                                    # the function that we want to make procedural
                                    # called us so we can't call to make it procedural
                                    # until it's done
                                    cf_eps[bva] = (startva, bflags)
                                else:
                                    logger.debug("descending into function 0x%x (from 0x%x)", bva, va)
                                    self.addEntryPoint(bva, arch=bflags)

                            if self._cf_noret.get(bva):
                                # then our next va is noflow!
                                self.addNoFlow(va, nextva)

                            calls_from[bva] = True

                            # We only go up to procedural branches, not across
                            continue

                    # we're jumping to a function we're in the middle of
                    # it's effectively a call from, but we should block
                    # until the other finishes processing to avoid some...odd
                    # issues with noret detection
                    if bva in self._cf_blocks and op.iflags & envi.IF_BRANCH:
                        if self._cf_recurse and startva != bva:
                            self._cf_delayed[startva].add(bva)
                            self._cf_delaying[bva].add(startva)

                        continue
                except Exception as e:
                    logger.warning("codeflow: %r", e, exc_info=True)

                if not opdone.get(bva):
                    optodo.append(((va, bva), bflags))

        # remove our local blocks from global block stack
        self._cf_blocks.pop()
        for fva, (pva, othrarch) in cf_eps.items():
            if fva in self._cf_blocks:
                self._cf_blocked[fva] = (pva, othrarch)
            else:
                if not self._mem.isFunction(fva):
                    self.addEntryPoint(fva, arch=othrarch)

        fallback = collections.OrderedDict()
        items = list(self._cf_blocked.items())
        for fva, othrarch in items:
            if fva not in self._cf_blocks and not self._mem.isFunction(fva):
                self._funcs.pop(fva, None)
                self._cf_blocked.pop(fva, None)
                self.addEntryPoint(fva, arch=othrarch)
            else:
                fallback[fva] = arch
        self._cf_blocked = fallback

        return list(calls_from.keys())

    def addEntryPoint(self, va, arch=envi.ARCH_DEFAULT):
        '''
        Analyze the given procedure entry point and flow downward
        to find all subsequent code blocks and procedure edges.

        Example:
            cf.addEntryPoint( 0x77c70308 )
            ... callbacks flow along ...
        '''
        # Architecture gets to decide on actual final VA and Architecture (ARM/THUMB/etc...)
        info = {'arch': arch}
        va, info = self._mem.arch.archModifyFuncAddr(va, info)
        arch = info.get('arch', arch)

        # Check if this is already a known function.
        if self._funcs.get(va) is not None:
            logger.debug("... skipping function 0x%x (arch: %x): already analyzed", va, arch)
            return

        # Add this function to known functions
        self._funcs[va] = True
        calls_from = self.addCodeFlow(va, arch=arch)
        self._fcalls[va] = calls_from

        # TODO: this ones spams. A lot.
        # logger.debug('addEntryPoint(0x%x): calls_from: %r', va, calls_from)

        # Finally, notify the callback of a new function
        # we gotta hold some of these off for a bit
        if va not in self._cf_delayed:
            self._cb_function(va, {'CallsFrom': calls_from})
            # remove this function from any blocking lists
            if va in self._cf_delaying:
                todo = []
                for blocked in self._cf_delaying[va]:
                    self._cf_delayed[blocked].discard(va)
                    if len(self._cf_delayed[blocked]) == 0:
                        todo.append(blocked)

                self._cf_delaying.pop(va, None)
                for ova in todo:
                    self._cf_delayed.pop(ova, None)
                    calls = self._calls_from.pop(ova, {})
                    self._cb_function(ova, {'CallsFrom': calls})
        else:
            # stash these off for later
            self._calls_from[va] = calls_from

        return va

    def flushFunction(self, fva):
        '''
        Codeflow context maintains a list of identified functions, to avoid
        analyzing the same function twice.  If a function is misidentified
        flushFunction() is used to clear that function from the tracked _funcs
        '''
        self._funcs[fva] = None

    def addDynamicBranchHandler(self, cb):
        '''
        Add a callback handler for dynamic branches the code-flow resolver
        doesn't know what to do with
        '''
        if cb in self._dynamic_branch_handlers:
            raise Exception("Already have this handler (%s) for dynamic branches" % repr(cb))

        self._dynamic_branch_handlers.append(cb)



import logging

logger = logging.getLogger(__name__)


def analyze(vw):

    for fname in vw.getFiles():

        h = vw.getFileMeta(fname, 'PELocalHints', None)
        if h is None:
            continue

        for fva, hints in h.items():
            if not vw.isFunction(fva):
                vw.makeFunction(fva)

            for name, offset, size, flags in hints:

                if offset > 0:
                    # Take 2 away for offset from frame....
                    offset = offset - (vw.psize * 2)
                    idx = offset / vw.psize

                    if idx > 100:
                        continue

                    logger.warning('FIXME: %d %s', offset, name)

                    #atype, aname = vw.getFunctionArg(fva, idx)
                    #if atype is None:
                        #atype = viv_magic.Unknown
                    #print 'ARG',idx, name
                    #vw.setFunctionArg(fva, idx, atype, name)

                elif offset < 0:
                    # "offset" is from frame, *we* are from initial esp...
                    offset -= vw.psize
                    vw.setFunctionLocal(fva, offset, None, name)



"""
The initial aarch64 module.
"""

import sys
import struct
import logging

import envi
import envi.bits as e_bits
from envi.const import *
from envi.archs.aarch64.regs import *
from envi.archs.aarch64 import A64Module

logger = logging.getLogger(__name__)

# CPU state (memory, regs inc SPSRs and banked registers)
# CPU mode  (User, FIQ, IRQ, supervisor, Abort, Undefined, System)
# 
# instruction code
# exception handler code

# calling conventions

## NO FREAKING CLUE YET....
class A64Call(envi.CallingConvention):
    arg_def = [(CC_REG, REG_X0), (CC_REG, REG_X1), (CC_REG, REG_X2), (CC_REG, REG_X3), 
            (CC_REG, REG_X4), (CC_REG, REG_X5), (CC_REG, REG_X6), (CC_REG, REG_X7), 
            (CC_STACK_INF, 8),]
    retaddr_def = (CC_REG, REG_X30)
    retval_def = (CC_REG, REG_X0)
    flags = CC_CALLEE_CLEANUP
    align = 8
    pad = 0

a64call = A64Call()
###


class CoProcEmulator:       # useful for prototyping, but should be subclassed
    def __init__(self):
        pass

    def stc(self, parms):
        logger.info("CoProcEmu: stc(%r)", parms)
    def ldc(self, parms):
        logger.info("CoProcEmu: ldc(%r)", parms)
    def cdp(self, parms):
        logger.info("CoProcEmu: cdp(%r)", parms)
    def mcr(self, parms):
        logger.info("CoProcEmu: mcr(%r)", parms)
    def mcrr(self, parms):
        logger.info("CoProcEmu: mcrr(%r)", parms)
    def mrc(self, parms):
        logger.info("CoProcEmu: mrc(%r)", parms)
    def mrrc(self, parms):
        logger.info("CoProcEmu: mrrc(%r)", parms)


def _getRegIdx(idx, mode):
    '''
    Currently cheating and simply supporting only one mode and associated registers.
    '''
    if idx >= MAX_REGS:
        return idx

    ridx = idx
    #ridx = idx + (mode*17)  # account for different banks of registers
    #ridx = reg_table[ridx]  # magic pointers allowing overlapping banks of registers
    return ridx


# Conditional handlers based in bitfield.  Used for handling bitfields in an array
def c0000(flags):
    return flags & 4

def c0001(flags):
    return flags & 4 == 0

def c0010(flags):
    return flags & 2

def c0011(flags):
    return flags & 2 == 0

def c0100(flags):
    return flags & 8

def c0101(flags):
    return flags & 8 == 0

def c0110(flags):
    return flags & 1

def c0111(flags):
    return flags & 1 == 0

def c1000(flags):
    return (flags & 6) == 2

def c1001(flags):
    return (flags & c) in (0, 4, 6) # C clear or Z set

def c1010(flags):
    return (flags & 9) in (0, 9)    # N == V

def c1011(flags):
    return (flags & 9) in (1, 8)    # N != V

def c1100(flags):
    return (flags & 4) == 0 and (flags & 9) in (0, 9)   # Z==0, N==V

def c1101(flags):
    return (flags & 4) or (flags & 9) in (1, 8)         # Z==1 or N!=V (basically, "not c1100")


conditionals = [
        c0000,
        c0001,
        c0010,
        c0011,
        c0100,
        c0101,
        c0110,
        c0111,
        c1000,
        c1001,
        c1010,
        c1011,
        c1100,
        c1101,
        ]

class A64Emulator(A64Module, A64RegisterContext, envi.Emulator):
    def __init__(self):
        A64Module.__init__(self)

        # FIXME: this should be None's, and added in for each real coproc... but this will work for now.
        self.coprocs = [CoProcEmulator() for x in range(16)]       

        seglist = [ (0,0xffffffff) for x in range(6) ]
        envi.Emulator.__init__(self, A64Module())

        A64RegisterContext.__init__(self)

        self.addCallingConvention("a64call", a64call)

    def undefFlags(self):
        """
        Used in PDE.
        A flag setting operation has resulted in un-defined value.  Set
        the flags to un-defined as well.
        """
        self.setRegister(REG_EFLAGS, None)

    def setFlag(self, which, state):
        flags = self.getCPSR()
        if state:
            flags |= which
        else:
            flags &= ~which
        self.setCPSR(flags)

    def getFlag(self, which):
        flags = self.getCPSR()
        if flags == None:
            raise envi.PDEUndefinedFlag(self)
        return bool(flags & which)

    def readMemValue(self, addr, size):
        bytes = self.readMemory(addr, size)
        if bytes == None:
            return None
        if len(bytes) != size:
            raise Exception("Read Gave Wrong Length At 0x%.8x (va: 0x%.8x wanted %d got %d)" % (self.getProgramCounter(),addr, size, len(bytes)))
        if size == 1:
            return struct.unpack(b"B", bytes)[0]
        elif size == 2:
            return struct.unpack(b"<H", bytes)[0]
        elif size == 4:
            return struct.unpack(b"<I", bytes)[0]
        elif size == 8:
            return struct.unpack(b"<Q", bytes)[0]

    def writeMemValue(self, addr, value, size):
        if size == 1:
            bytes = struct.pack(b"B",value & 0xff)
        elif size == 2:
            bytes = struct.pack(b"<H",value & 0xffff)
        elif size == 4:
            bytes = struct.pack(b"<I", value & 0xffffffff)
        elif size == 8:
            bytes = struct.pack(b"<Q", value & 0xffffffffffffffff)
        self.writeMemory(addr, bytes)

    def readMemSignedValue(self, addr, size):
        bytes = self.readMemory(addr, size)
        if bytes == None:
            return None
        if size == 1:
            return struct.unpack(b"b", bytes)[0]
        elif size == 2:
            return struct.unpack(b"<h", bytes)[0]
        elif size == 4:
            return struct.unpack(b"<l", bytes)[0]

    def executeOpcode(self, op):
        # NOTE: If an opcode method returns
        #       other than None, that is the new eip
        x = None
        meth = self.op_methods.get(op.mnem, None)
        if meth == None:
            raise envi.UnsupportedInstruction(self, op)
        x = meth(op)

        if x == None:
            pc = self.getProgramCounter()
            x = pc+op.size

        self.setProgramCounter(x)

    def doPush(self, val):
        sp = self.getRegister(REG_SP)
        sp -= 8
        self.writeMemValue(sp, val, 8)
        self.setRegister(REG_SP, sp)

    def doPop(self):
        sp = self.getRegister(REG_SP)
        val = self.readMemValue(sp, 8)
        self.setRegister(REG_SP, sp+8)
        return val

    def getProcMode(self):
        '''
        get current AARCH64 processor mode.  see proc_modes (const.py)
        '''
        return self._rctx_vals[REG_CPSR] & 0x1f     # obfuscated for speed.  could call getCPSR but it's not as fast

    def getCPSR(self):
        '''
        return the Current Program Status Register.
        '''
        return self._rctx_vals[REG_CPSR]

    def setCPSR(self, psr, mask=0xffffffff):
        '''
        set the CPSR for the current AARCH64 processor mode
        '''
        psr = self._rctx_vals[REG_CPSR] & (~mask) | (psr & mask)
        self._rctx_vals[REG_CPSR] = psr

    def getSPSR(self, mode):
        '''
        get the SPSR for the given AARCH64 processor mode
        '''
        ridx = _getRegIdx(REG_SPSR, mode)
        return self._rctx_vals[ridx]

    def setSPSR(self, mode, psr, mask=0xffffffff):
        '''
        set the SPSR for the given AARCH64 processor mode
        '''
        ridx = _getRegIdx(PSR_Offset, mode)
        psr = self._rctx_vals[REG_CPSR] & (~mask) | (psr & mask)
        self._rctx_vals[ridx] = psr

    def setProcMode(self, mode):
        '''
        set the current processor mode.  stored in CPSR
        '''
        # write current psr to the saved psr register for target mode
        # but not for USR or SYS modes, which don't have their own SPSR
        if mode not in (PM_usr, PM_sys):
            curSPSRidx = proc_modes[mode]
            self._rctx_vals[curSPSRidx] = self.getCPSR()

        # set current processor mode
        cpsr = self._rctx_vals[REG_CPSR] & 0xffffffe0
        self._rctx_vals[REG_CPSR] = cpsr | mode

    def getRegister(self, index, mode=None):
        """
        Return the current value of the specified register index.
        """
        if mode == None:
            mode = self.getProcMode() & 0xf
        else:
            mode &= 0xf

        idx = (index & 0xffff)
        ridx = _getRegIdx(idx, mode)
        if idx == index:
            return self._rctx_vals[ridx]

        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = (2**width)-1
        return (self._rctx_vals[ridx] >> offset) & mask

    def setRegister(self, index, value, mode=None):
        """
        Set a register value by index.
        """
        if mode == None:
            mode = self.getProcMode() & 0xf
        else:
            mode &= 0xf

        self._rctx_dirty = True

        idx = (index & 0xffff)
        ridx = _getRegIdx(idx, mode)

        if idx == index:    # not a metaregister
            self._rctx_vals[ridx] = (value & self._rctx_masks[ridx])
            return

        # If we get here, it's a meta register index.
        # NOTE: offset/width are in bits...
        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = e_bits.b_masks[width]
        mask = mask << offset

        # NOTE: basewidth is in *bits*
        basewidth = self._rctx_widths[ridx]
        basemask  = (2**basewidth)-1

        # cut a whole in basemask at the size/offset of mask
        finalmask = basemask ^ mask

        curval = self._rctx_vals[ridx]

        self._rctx_vals[ridx] = (curval & finalmask) | (value << offset)

    def integerSubtraction(self, op):
        """
        Do the core of integer subtraction but only *return* the
        resulting value rather than assigning it.
        (allows cmp and sub to use the same code)
        """
        # Src op gets sign extended to dst
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        Sflag = op.iflags & IF_PSR_S

        if src1 == None or src2 == None:
            self.undefFlags()
            return None

        return self.intSubBase(src1, src2, Sflag)

    def intSubBase(self, src1, src2, Sflag=0, rd=0):
        # So we can either do a BUNCH of crazyness with xor and shifting to
        # get the necessary flags here, *or* we can just do both a signed and
        # unsigned sub and use the results.


        udst = e_bits.unsigned(src1, 4)
        usrc = e_bits.unsigned(src2, 4)

        sdst = e_bits.signed(src1, 4)
        ssrc = e_bits.signed(src2, 4)

        ures = udst - usrc
        sres = sdst - ssrc

        if Sflag:
            curmode = self.getProcMode() 
            if rd == 15:
                if(curmode != PM_sys and curmode != PM_usr):
                    self.setCPSR(self.getSPSR(curmode))
                else:
                    raise Exception("Messed up opcode...  adding to r15 from PM_usr or PM_sys")
            self.setFlag(PSR_N_bit, e_bits.is_signed(ures, 4))
            self.setFlag(PSR_Z_bit, not ures)
            self.setFlag(PSR_C_bit, not e_bits.is_unsigned_carry(ures, 4))
            self.setFlag(PSR_V_bit, e_bits.is_signed_overflow(sres, 4))

        return ures

    def logicalAnd(self, op):
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)

        # PDE
        if src1 == None or src2 == None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        res = src1 & src2

        self.setFlag(PSR_N_bit, 0)
        self.setFlag(PSR_Z_bit, not res)
        self.setFlag(PSR_C_bit, 0)
        self.setFlag(PSR_V_bit, 0)
        return res

    def i_nop(self, op):
        return
        
    def i_and(self, op):
        res = self.logicalAnd(op)
        self.setOperValue(op, 0, res)
        
    def i_stp(self, op):
        dsize = op.opers[0].tsize
        val1 = self.getOperValue(op, 0)
        val2 = self.getOperValue(op, 1)
        baseaddr = self.getOperAddr(op, 2)
        self.writeMemValue(baseaddr, val1, dsize)
        self.writeMemValue(baseaddr + dsize, val2, dsize)


    def i_ldp(self, op):
        dsize = op.opers[0].tsize
        baseaddr = self.getOperAddr(op, 2)
        val1 = self.readMemValue(baseaddr, dsize)
        val2 = self.readMemValue(baseaddr + dsize, dsize)

        if op.opers[0].reg == REG_PC:
            return val1
        if op.opers[1].reg == REG_PC:
            return val2

    def i_ldr(self, op):
        # hint: covers ldr, ldrb, ldrbt, ldrd, ldrh, ldrsh, ldrsb, ldrt   (any instr where the syntax is ldr{condition}stuff)
        val = self.getOperValue(op, 1)
        self.setOperValue(op, 0, val)
        if op.opers[0].reg == REG_PC:
            return val

    def i_mov(self, op):
        val = self.getOperValue(op, 1)
        self.setOperValue(op, 0, val)

    def i_movi(self, op):
        val = self.getOperValue(op, 1)
        self.setOperValue(op, 0, val)

    i_msr = i_mov
    i_adr = i_mov
    i_adrp = i_mov
    # confirm these next ones are ok
    i_movk = i_mov
    i_movz = i_mov

    def i_rev(self, op):
        # rev
        val = self.getOperValue(op, 0)
        self.setOperValue(op, 1, val)

    def i_str(self, op):
        # hint: covers str, strb, strbt, strd, strh, strsh, strsb, strt   (any instr where the syntax is str{condition}stuff)
        val = self.getOperValue(op, 0)
        self.setOperValue(op, 1, val)

    def i_add(self, op):
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        
        if src1 == None or src2 == None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        dsize = op.opers[0].tsize
        ssize = op.opers[1].tsize
        s2size = op.opers[2].tsize

        usrc1 = e_bits.unsigned(src1, 4)
        usrc2 = e_bits.unsigned(src2, 4)
        ssrc1 = e_bits.signed(src1, 4)
        ssrc2 = e_bits.signed(src2, 4)

        ures = usrc1 + usrc2
        sres = ssrc1 + ssrc2


        self.setOperValue(op, 0, ures)

        curmode = self.getProcMode() 
        if op.iflags & IF_S:
            if op.opers[0].reg == 15 and (curmode != PM_sys and curmode != PM_usr):
                self.setCPSR(self.getSPSR(curmode))
            else:
                raise Exception("Messed up opcode...  adding to r15 from PM_usr or PM_sys")
            self.setFlag(PSR_N_bit, e_bits.is_signed(ures, dsize))
            self.setFlag(PSR_Z_bit, not ures)
            self.setFlag(PSR_C_bit, e_bits.is_unsigned_carry(ures, dsize))
            self.setFlag(PSR_V_bit, e_bits.is_signed_overflow(sres, dsize))

    def i_b(self, op):
        '''
        conditional branches (eg. bne) will be handled here
        '''
        return self.getOperValue(op, 0)

    i_br = i_b

    def i_bl(self, op):
        self.setRegister(REG_LR, self.getRegister(REG_PC))
        return self.getOperValue(op, 0)

    def i_tst(self, op):
        src1 = self.getOperValue(op, 0)
        src2 = self.getOperValue(op, 1)

        dsize = op.opers[0].tsize
        ures = src1 & src2

        self.setFlag(PSR_N_bit, e_bits.is_signed(ures, dsize))
        self.setFlag(PSR_Z_bit, (0,1)[ures==0])
        self.setFlag(PSR_C_bit, e_bits.is_unsigned_carry(ures, dsize))
        
    def i_rsb(self, op):
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        
        if src1 == None or src2 == None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        dsize = op.opers[0].tsize
        ssize = op.opers[1].tsize
        s2size = op.opers[2].tsize

        usrc1 = e_bits.unsigned(src1, 4)
        usrc2 = e_bits.unsigned(src2, 4)
        ssrc1 = e_bits.signed(src1, 4)
        ssrc2 = e_bits.signed(src2, 4)

        ures = usrc2 - usrc1
        sres = ssrc2 - ssrc1


        self.setOperValue(op, 0, ures)

        curmode = self.getProcMode() 
        if op.iflags & IF_S:
            if op.opers[0].reg == 15:
                if (curmode != PM_sys and curmode != PM_usr):
                    self.setCPSR(self.getSPSR(curmode))
                else:
                    raise Exception("Messed up opcode...  adding to r15 from PM_usr or PM_sys")
            self.setFlag(PSR_C_bit, e_bits.is_unsigned_carry(ures, dsize))
            self.setFlag(PSR_Z_bit, not ures)
            self.setFlag(PSR_N_bit, e_bits.is_signed(ures, dsize))
            self.setFlag(PSR_V_bit, e_bits.is_signed_overflow(sres, dsize))

    def i_rsb(self, op):
        # Src op gets sign extended to dst
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        Sflag = op.iflags & IF_PSR_S

        if src1 == None or src2 == None:
            self.undefFlags()
            return None

        res = self.intSubBase(src2, src1, Sflag, op.opers[0].reg)
        self.setOperValue(op, 0, res)

    def i_sub(self, op):
        # Src op gets sign extended to dst
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        Sflag = op.iflags & IF_PSR_S

        if src1 == None or src2 == None:
            self.undefFlags()
            return None

        res = self.intSubBase(src1, src2, Sflag, op.opers[0].reg)
        self.setOperValue(op, 0, res)

    i_subs = i_sub

    def i_eor(self, op):
        src1 = self.getOperValue(op, 1)
        src2 = self.getOperValue(op, 2)
        
        if src1 == None or src2 == None:
            self.undefFlags()
            self.setOperValue(op, 0, None)
            return

        usrc1 = e_bits.unsigned(src1, 4)
        usrc2 = e_bits.unsigned(src2, 4)

        ures = usrc1 ^ usrc2

        self.setOperValue(op, 0, ures)

        curmode = self.getProcMode() 
        if op.iflags & IF_S:
            if op.opers[0].reg == 15:
                if (curmode != PM_sys and curmode != PM_usr):
                    self.setCPSR(self.getSPSR(curmode))
                else:
                    raise Exception("Messed up opcode...  adding to r15 from PM_usr or PM_sys")
            self.setFlag(PSR_C_bit, e_bits.is_unsigned_carry(ures, 4))
            self.setFlag(PSR_Z_bit, not ures)
            self.setFlag(PSR_N_bit, e_bits.is_signed(ures, 4))
            self.setFlag(PSR_V_bit, e_bits.is_signed_overflow(sres, 4))

    def i_cmp(self, op):
        # Src op gets sign extended to dst
        src1 = self.getOperValue(op, 0)
        src2 = self.getOperValue(op, 1)
        Sflag = op.iflags & IF_PSR_S

        res = self.intSubBase(src1, src2, Sflag, op.opers[0].reg)

    i_cmps = i_cmp


    # Coprocessor Instructions
    def i_stc(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.stc(op.opers)

    def i_ldc(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.ldc(op.opers)

    def i_cdp(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.cdp(op.opers)

    def i_mrc(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.mrc(op.opers)

    def i_mrrc(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.mrrc(op.opers)

    def i_mcr(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.mrrc(op.opers)

    def i_mcrr(self, op):
        cpnum = op.opers[0]
        coproc = self._getCoProc(cpnum)
        coproc.mcrr(op.opers)


